from .batch_filter import BatchFilter
from gunpowder.array import Array
import collections
import itertools
import logging
import numpy as np

logger = logging.getLogger(__name__)


class BalanceLabels(BatchFilter):
    '''Creates a scale array to balance the loss between class labels.

    Note that this only balances loss weights per-batch and does not accumulate
    statistics about class balance across batches.

    Args:

        labels (:class:`ArrayKey`):

            An array containing binary or integer labels.

        scales (:class:`ArrayKey`):

            A array with scales to be created. This new array will have the
            same ROI and resolution as ``labels``.

        mask (:class:`ArrayKey`, optional):

            An optional mask (or list of masks) to consider for balancing.
            Every voxel marked with a 0 will not contribute to the scaling and
            will have a scale of 0 in ``scales``.

        slab (``tuple`` of ``int``, optional):

            A shape specification to perform the balancing in slabs of this
            size. -1 can be used to refer to the actual size of the label
            array. For example, a slab of::

                (2, -1, -1, -1)

            will perform the balancing for every each slice ``[0:2,:]``,
            ``[2:4,:]``, ... individually.

        num_classes(``int``, optional):

            The number of classes. Labels will be expected to be in the
            interval [0, ``num_classes``). Defaults to 2 for binary
            classification.
    '''

    def __init__(self, labels, scales, mask=None, slab=None, num_classes=2):

        self.labels = labels
        self.scales = scales
        if mask is None:
            self.masks = []
        elif not isinstance(mask, collections.Iterable):
            self.masks = [mask]
        else:
            self.masks = mask

        self.slab = slab
        self.num_classes = num_classes

    def setup(self):

        assert self.labels in self.spec, (
            "Asked to balance labels %s, which are not provided."%self.labels)

        for mask in self.masks:
            assert mask in self.spec, (
                "Asked to apply mask %s to balance labels, but mask is not "
                "provided."%mask)

        spec = self.spec[self.labels].copy()
        spec.dtype = np.float32
        self.provides(self.scales, spec)
        self.enable_autoskip()

    def process(self, batch, request):

        labels = batch.arrays[self.labels]

        assert len(np.unique(labels.data)) <= self.num_classes, (
            "Found more unique labels than classes in %s."%self.labels)
        assert 0 <= np.min(labels.data) < self.num_classes, (
            "Labels %s are not in [0, num_classes)."%self.labels)
        assert 0 <= np.max(labels.data) < self.num_classes, (
            "Labels %s are not in [0, num_classes)."%self.labels)

        # initialize error scale with 1s
        error_scale = np.ones(labels.data.shape, dtype=np.float32)

        # set error_scale to 0 in masked-out areas
        for key in self.masks:
            mask = batch.arrays[key]
            assert labels.data.shape == mask.data.shape, (
                "Shape of mask %s %s does not match %s %s"%(
                    mask,
                    mask.data.shape,
                    self.labels,
                    labels.data.shape))
            error_scale *= mask.data

        if not self.slab:
            slab = error_scale.shape
        else:
            # slab with -1 replaced by shape
            slab = tuple(
                m if s == -1 else s
                for m, s in zip(error_scale.shape, self.slab))

        slab_ranges = (
            range(0, m, s)
            for m, s in zip(error_scale.shape, slab))

        for start in itertools.product(*slab_ranges):
            slices = tuple(
                slice(start[d], start[d] + slab[d])
                for d in range(len(slab)))
            self.__balance(
                labels.data[slices],
                error_scale[slices])

        spec = self.spec[self.scales].copy()
        spec.roi = labels.spec.roi
        batch.arrays[self.scales] = Array(error_scale, spec)

    def __balance(self, labels, scale):

        # in the masked-in area, compute the fraction of per-class samples
        masked_in = scale.sum()
        classes, counts = np.unique(labels[np.nonzero(scale)], return_counts=True)
        fracs = counts.astype(float) / masked_in if masked_in > 0 else np.zeros(counts.size)
        np.clip(fracs, 0.05, 0.95, fracs)

        # compute the class weights
        w_sparse = 1.0 / float(self.num_classes) / fracs
        w = np.zeros(self.num_classes)
        w[classes] = w_sparse

        # scale the masked-in scale with the class weights
        scale *= np.take(w, labels)



from __future__ import print_function
import logging
import traceback

import sys

logger = logging.getLogger(__name__)


class NoSuchModule(object):
    def __init__(self, name):
        self.__name = name
        self.__traceback_str = traceback.format_tb(sys.exc_info()[2])
        errtype, value = sys.exc_info()[:2]
        self.__exception = errtype(value)

    def __getattr__(self, item):
        raise self.__exception

try:
    import dvision
except ImportError as e:
    dvision = NoSuchModule('dvision')

try:
    import h5py
except ImportError as e:
    h5py = NoSuchModule('h5py')

try:
    import pyklb
except ImportError as e:
    pyklb = NoSuchModule('pyklb')

try:
    import caffe
except ImportError as e:
    caffe = NoSuchModule('caffe')

try:
    import tensorflow
except ImportError as e:
    tensorflow = NoSuchModule('tensorflow')

try:
    import malis
except ImportError as e:
    malis = NoSuchModule('malis')

try:
    import augment
except ImportError as e:
    augment = NoSuchModule('augment')

try:
    import z5py
except ImportError as e:
    z5py = NoSuchModule('z5py')

try:
    import zarr
    from .zarr_file import ZarrFile
except ImportError as e:
    zarr = NoSuchModule('zarr')
    ZarrFile = None



import copy
import logging

from .batch_provider import BatchProvider
from gunpowder.profiling import Timing

logger = logging.getLogger(__name__)

class BatchFilter(BatchProvider):
    '''Convenience wrapper for :class:`BatchProviders<BatchProvider>` with
    exactly one input provider.

    By default, a node of this class will expose the same :class:`ProviderSpec`
    as the upstream provider. You can modify the provider spec by calling
    :func:`provides` and :func:`updates` in :func:`setup`.

    Subclasses need to implement at least :func:`process` to modify a passed
    batch (downstream). Optionally, the following methods can be implemented:

        :func:`setup`

            Initialize this filter. Called after setup of the DAG. All upstream 
            providers will be set up already.

        :func:`teardown`

            Destruct this filter, free resources, stop worker processes.

        :func:`prepare`

            Prepare for a batch request. Always called before each 
            :func:`process`. Use it to modify a batch request to be passed
            upstream.
    '''

    def get_upstream_provider(self):
        assert len(self.get_upstream_providers()) == 1, "BatchFilters need to have exactly one upstream provider"
        return self.get_upstream_providers()[0]

    def updates(self, key, spec):
        '''Update an output provided by this :class:`BatchFilter`.

        Implementations should call this in their :func:`setup` method, which
        will be called when the pipeline is build.

        Args:

            key (:class:`ArrayKey` or :class:`PointsKey`):

                The array or point set key this filter updates.

            spec (:class:`ArraySpec` or :class:`PointsSpec`):

                The updated spec of the array or point set.
        '''

        assert key in self.spec, "Node %s is trying to change the spec for %s, but is not provided upstream."%(type(self).__name__, key)
        self.spec[key] = copy.deepcopy(spec)
        self.updated_items.append(key)

        logger.debug("%s updates %s with %s"%(self.name(), key, spec))

    def enable_autoskip(self, skip=True):
        '''Enable automatic skipping of this :class:`BatchFilter`, based on
        given :func:`updates` and :func:`provides` calls. Has to be called in
        :func:`setup`.

        By default, :class:`BatchFilters<BatchFilter>` are not skipped
        automatically, regardless of what they update or provide. If autskip is
        enabled, :class:`BatchFilters<BatchFilter>` will only be run if the
        request contains at least one key reported earlier with
        :func:`updates` or :func:`provides`.
        '''

        self._autoskip_enabled = skip

    def _init_spec(self):
        # default for BatchFilters is to provide the same as upstream
        if not hasattr(self, '_spec') or self._spec is None:
            self._spec = copy.deepcopy(self.get_upstream_provider().spec)

    def internal_teardown(self):

        logger.debug("Resetting spec of %s", self.name())
        self._spec = None
        self._updated_items = []

        self.teardown()

    @property
    def updated_items(self):
        '''Get a list of the keys that are updated by this `BatchFilter`.

        This list is only available after the pipeline has been build. Before
        that, it is empty.
        '''

        if not hasattr(self, '_updated_items'):
            self._updated_items = []

        return self._updated_items

    @property
    def autoskip_enabled(self):

        if not hasattr(self, '_autoskip_enabled'):
            self._autoskip_enabled = False

        return self._autoskip_enabled

    def provide(self, request):

        # operate on a copy of the request, to provide the original request to
        # 'process' for convenience
        upstream_request = copy.deepcopy(request)

        skip = self.__can_skip(request)

        timing_prepare = Timing(self, 'prepare')
        timing_prepare.start()

        if not skip:
            self.prepare(upstream_request)
            self.remove_provided(upstream_request)

        timing_prepare.stop()

        batch = self.get_upstream_provider().request_batch(upstream_request)

        timing_process = Timing(self, 'process')
        timing_process.start()

        if not skip:
            self.process(batch, request)

        timing_process.stop()

        batch.profiling_stats.add(timing_prepare)
        batch.profiling_stats.add(timing_process)

        return batch

    def __can_skip(self, request):
        '''Check if this filter needs to be run for the given request.'''

        if not self.autoskip_enabled:
            return False

        for key, _ in request.items():
            if key in self.provided_items:
                return False
            if key in self.updated_items:
                return False

        return True

    def setup(self):
        '''To be implemented in subclasses.

        Called during initialization of the DAG. Callees can assume that all
        upstream providers are set up already.

        In setup, call :func:`provides` or :func:`updates` to announce the
        arrays and points provided or changed by this node.
        '''
        pass

    def prepare(self, request):
        '''To be implemented in subclasses.

        Prepare for a batch request. Change the request as needed, it will be
        passed on upstream.
        '''
        pass

    def process(self, batch, request):
        '''To be implemented in subclasses.

        Filter a batch, will be called after :func:`prepare`. Change batch as
        needed, it will be passed downstream. ``request`` is the same as passed
        to :func:`prepare`, provided for convenience.

        Args:

            batch (:class:`Batch`):

                The batch received from upstream to be modified by this node.

            request (:class:`BatchRequest`):

                The request this node received. The updated batch should meet
                this request.
        '''
        raise RuntimeError("Class %s does not implement 'process'"%type(self).__name__)



import logging
import multiprocessing

from .freezable import Freezable
from .profiling import ProfilingStats
from .array import Array, ArrayKey
from .points import Points, PointsKey

logger = logging.getLogger(__name__)

class Batch(Freezable):
    '''Contains the requested batch as a collection of :class:`Arrays<Array>`
    and :class:`Points` that is passed through the pipeline from sources to
    sinks.

    This collection mimics a dictionary. Items can be added with::

        batch = Batch()
        batch[array_key] = Array(...)
        batch[points_key] = Points(...)

    Here, ``array_key`` and ``points_key`` are :class:`ArrayKey` and
    :class:`PointsKey`. The items can be queried with::

        array = batch[array_key]
        points = batch[points_key]

    Furthermore, pairs of keys/values can be iterated over using
    ``batch.items()``.

    To access only arrays or point sets, use the dictionaries ``batch.arrays``
    or ``batch.points``, respectively.

    Attributes:

        arrays (dict from :class:`ArrayKey` to :class:`Array`):

            Contains all arrays that have been requested for this batch.

        points (dict from :class:`PointsKey` to :class:`Points`):

            Contains all point sets that have been requested for this batch.
    '''

    __next_id = multiprocessing.Value('L')

    @staticmethod
    def get_next_id():
        with Batch.__next_id.get_lock():
            next_id = Batch.__next_id.value
            Batch.__next_id.value += 1
        return next_id

    def __init__(self):

        self.id = Batch.get_next_id()
        self.profiling_stats = ProfilingStats()
        self.arrays = {}
        self.points  = {}
        self.affinity_neighborhood = None
        self.loss = None
        self.iteration = None

        self.freeze()

    def __setitem__(self, key, value):

        if isinstance(value, Array):
            assert isinstance(key, ArrayKey), (
                "Only a ArrayKey is allowed as key for an Array value.")
            self.arrays[key] = value

        elif isinstance(value, Points):
            assert isinstance(key, PointsKey), (
                "Only a PointsKey is allowed as key for a Points value.")
            self.points[key] = value

        else:
            raise RuntimeError(
                "Only Array or Points can be set in a %s."%type(self).__name__)

    def __getitem__(self, key):

        if isinstance(key, ArrayKey):
            return self.arrays[key]

        elif isinstance(key, PointsKey):
            return self.points[key]

        else:
            raise RuntimeError(
                "Only ArrayKey or PointsKey can be used as keys in a "
                "%s."%type(self).__name__)

    def __len__(self):

        return len(self.arrays) + len(self.points)

    def __contains__(self, key):

        if isinstance(key, ArrayKey):
            return key in self.arrays

        elif isinstance(key, PointsKey):
            return key in self.points

        else:
            raise RuntimeError(
                "Only ArrayKey or PointsKey can be used as keys in a "
                "%s."%type(self).__name__)

    def __delitem__(self, key):

        if isinstance(key, ArrayKey):
            del self.arrays[key]

        elif isinstance(key, PointsKey):
            del self.points[key]

        else:
            raise RuntimeError(
                "Only ArrayKey or PointsKey can be used as keys in a "
                "%s."%type(self).__name__)

    def items(self):
        '''Provides a generator iterating over key/value pairs.'''

        for (k, v) in self.arrays.items():
            yield k, v
        for (k, v) in self.points.items():
            yield k, v

    def get_total_roi(self):
        '''Get the union of all the array ROIs in the batch.'''

        total_roi = None

        for collection_type in [self.arrays, self.points]:
            for (key, obj) in collection_type.items():
                if total_roi is None:
                    total_roi = obj.spec.roi
                else:
                    total_roi = total_roi.union(obj.spec.roi)

        return total_roi

    def __repr__(self):

        r = ""
        for collection_type in [self.arrays, self.points]:
            for (key, obj) in collection_type.items():
                r += "%s: %s\n"%(key, obj.spec)
        return r



import logging
import numpy as np

from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)

class Normalize(BatchFilter):
    '''Normalize the values of an array to be floats between 0 and 1, based on
    the type of the array.

    Args:

        array (:class:`ArrayKey`):

            The key of the array to modify.

        factor (scalar, optional):

            The factor to use. If not given, a factor is chosen based on the
            ``dtype`` of the array (e.g., ``np.uint8`` would result in a factor
            of ``1.0/255``).

        dtype (data-type, optional):

            The datatype of the normalized array. Defaults to ``np.float32``.
    '''

    def __init__(self, array, factor=None, dtype=np.float32):

        self.array = array
        self.factor = factor
        self.dtype = dtype

    def process(self, batch, request):

        if self.array not in batch.arrays:
            return

        factor = self.factor
        array = batch.arrays[self.array]

        if factor is None:

            logger.debug("automatically normalizing %s with dtype=%s",
                    self.array, array.data.dtype)

            if array.data.dtype == np.uint8:
                factor = 1.0/255
            elif array.data.dtype == np.uint16:
                factor = 1.0/(255*255)
            elif array.data.dtype == np.float32:
                assert array.data.min() >= 0 and array.data.max() <= 1, (
                        "Values are float but not in [0,1], I don't know how "
                        "to normalize. Please provide a factor.")
                factor = 1.0
            else:
                raise RuntimeError("Automatic normalization for " +
                        str(array.data.dtype) + " not implemented, please "
                        "provide a factor.")

        logger.debug("scaling %s with %f", self.array, factor)
        array.data = array.data.astype(self.dtype)*factor



import distutils.util
import numpy as np
import logging
import requests
from copy import deepcopy

from gunpowder.batch import Batch
from gunpowder.coordinate import Coordinate
from gunpowder.ext import dvision
from gunpowder.nodes.batch_provider import BatchProvider
from gunpowder.points import PointsKeys, Points
from gunpowder.profiling import Timing
from gunpowder.roi import Roi

from gunpowder.contrib.points import PreSynPoint, PostSynPoint

logger = logging.getLogger(__name__)

class DvidPartnerAnnoationSourceReadException(Exception):
    pass

class MaskNotProvidedException(Exception):
    pass


# TODO: This seems broken. There is code involving a voxel size, but points
# don't have voxel sizes
class DvidPartnerAnnotationSource(BatchProvider):
    '''
        :param hostname: hostname for DVID server
        :type hostname: str

        :param port: port for DVID server
        :type port: int

        :param uuid: UUID of node on DVID server
        :type uuid: str

        :param datasets: dict {PointsKey: DVID data instance}
    '''

    def __init__(
            self,
            hostname,
            port,
            uuid,
            datasets=None,
            rois=None):

        self.hostname = hostname
        self.port = port
        self.url = "http://{}:{}".format(self.hostname, self.port)
        self.uuid = uuid

        self.datasets = datasets if datasets is not None else {}
        self.rois = rois if rois is not None else {}

        self.node_service = None
        self.dims = 0

    def setup(self):

        for points_key, points_name in self.datasets.items():
            self.provides(
                points_key,
                PointsSpec(roi=self.points_rois[points_key]))

        logger.info("DvidPartnerAnnotationSource.spec:\n{}".format(self.spec))

    def provide(self, request):

        timing = Timing(self)
        timing.start()

        batch = Batch()

        # if pre and postsynaptic locations requested, their id : SynapseLocation dictionaries should be created
        # together s.t. the ids are unique and allow to find partner locations
        if PointsKeys.PRESYN in request.points or PointsKeys.POSTSYN in request.points:
            try:  # either both have the same roi, or only one of them is requested
                assert request.points[PointsKeys.PRESYN] == request.points[PointsKeys.POSTSYN]
            except:
                assert PointsKeys.PRESYN not in request.points or PointsKeys.POSTSYN not in request.points
            if PointsKeys.PRESYN in request.points:
                presyn_points, postsyn_points = self.__read_syn_points(roi=request.points[PointsKeys.PRESYN])
            elif PointsKeys.POSTSYN in request.points:
                presyn_points, postsyn_points = self.__read_syn_points(roi=request.points[PointsKeys.POSTSYN])

        for (points_key, roi) in request.points.items():
            # check if requested points can be provided
            if points_key not in self.spec:
                raise RuntimeError("Asked for %s which this source does not provide"%points_key)
            # check if request roi lies within provided roi
            if not self.spec[points_key].roi.contains(roi):
                raise RuntimeError("%s's ROI %s outside of my ROI %s"%(points_key,roi,self.spec[points_key].roi))

            logger.debug("Reading %s in %s..."%(points_key, roi))
            id_to_point = {PointsKeys.PRESYN: presyn_points,
                           PointsKeys.POSTSYN: postsyn_points}[points_key]

            batch.points[points_key] = Points(
                data=id_to_point,
                spec=PointsSpec(roi=roi))

        logger.debug("done")

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch

    def __load_json_annotations(self, array_shape_voxel, array_offset_voxel, array_name):
        url = "http://" + str(self.hostname) + ":" + str(self.port)+"/api/node/" + str(self.uuid) + '/' + \
              str(array_name) + "/elements/{}_{}_{}/{}_{}_{}".format(array_shape_voxel[2], array_shape_voxel[1], array_shape_voxel[0],
                                                   array_offset_voxel[2], array_offset_voxel[1], array_offset_voxel[0])
        annotations_file = requests.get(url)
        json_annotations = annotations_file.json()
        if json_annotations is None:
            json_annotations = []  # create empty_dummy_json_annotations
            # raise Exception ('No synapses found in region defined by array_offset {} and array_shape {}'.format(array_offset, array_shape))
        return json_annotations

    def __read_syn_points(self, roi):
        """ read json file from dvid source, in json format to create a PreSynPoint/PostSynPoint for every location given """

        if PointsKeys.PRESYN in self.points_voxel_size:
            voxel_size = self.points_voxel_size[PointsKeys.PRESYN]
        elif PointsKeys.POSTSYN in self.points_voxel_size:
            voxel_size = self.points_voxel_size[PointsKeys.POSTSYN]

        syn_file_json = self.__load_json_annotations(array_shape_voxel  = roi.get_shape() // voxel_size,
                                                     array_offset_voxel = roi.get_offset() // voxel_size,
                                                     array_name    = self.datasets[PointsKeys.PRESYN])

        presyn_points_dict, postsyn_points_dict = {}, {}
        location_to_location_id_dict, location_id_to_partner_locations = {}, {}
        for node_nr, node in enumerate(syn_file_json):
            # collect information
            kind        = str(node['Kind'])
            location    = np.asarray((node['Pos'][2], node['Pos'][1], node['Pos'][0])) * voxel_size
            location_id = int(node_nr)
            # some synapses are wrongly annotated in dvid source, have 'Tag': null ???, they are skipped
            try:
                syn_id = int(node['Tags'][0][3:])
            except:
                continue
            location_to_location_id_dict[str(location)] = location_id

            partner_locations = []
            try:
                for relation in node['Rels']:
                    partner_locations.append((np.asarray([relation['To'][2], relation['To'][1], relation['To'][0]]))*voxel_size)
            except:
                partner_locations = []
            location_id_to_partner_locations[int(node_nr)] = partner_locations

            # check if property given, not always given
            props = {}
            if 'conf' in node['Prop']:
                props['conf'] = float(node['Prop']['conf'])
            if 'agent' in node['Prop']:
                props['agent']  = str(node['Prop']['agent'])
            if 'flagged' in node['Prop']:
                str_value_flagged = str(node['Prop']['flagged'])
                props['flagged']  = bool(distutils.util.strtobool(str_value_flagged))
            if 'multi' in node['Prop']:
                str_value_multi = str(node['Prop']['multi'])
                props['multi']  = bool(distutils.util.strtobool(str_value_multi))

            # create synPoint with information collected so far (partner_ids not completed yet)
            if kind == 'PreSyn':
                syn_point = PreSynPoint(location=location, location_id=location_id,
                                     synapse_id=syn_id, partner_ids=[], props=props)
                presyn_points_dict[int(node_nr)] = deepcopy(syn_point)
            elif kind == 'PostSyn':
                syn_point = PostSynPoint(location=location, location_id=location_id,
                                     synapse_id=syn_id, partner_ids=[], props=props)
                postsyn_points_dict[int(node_nr)] = deepcopy(syn_point)

        # add partner ids
        last_node_nr = len(syn_file_json)-1
        for current_syn_point_id in location_id_to_partner_locations.keys():
            all_partner_ids = []
            for partner_loc in location_id_to_partner_locations[current_syn_point_id]:
                if location_to_location_id_dict.has_key(str(partner_loc)):
                    all_partner_ids.append(int(location_to_location_id_dict[str(partner_loc)]))
                else:
                    last_node_nr = last_node_nr + 1
                    assert not location_to_location_id_dict.has_key(str(partner_loc))
                    all_partner_ids.append(int(last_node_nr))

            if current_syn_point_id in presyn_points_dict:
                presyn_points_dict[current_syn_point_id].partner_ids = all_partner_ids
            elif current_syn_point_id in postsyn_points_dict:
                postsyn_points_dict[current_syn_point_id].partner_ids = all_partner_ids
            else:
                raise Exception("current syn_point id not found in any dictionary")

        return presyn_points_dict, postsyn_points_dict


    def __repr__(self):
        return "DvidPartnerAnnoationSource(hostname={}, port={}, uuid={}, raw_array_name={}, gt_array_name={}".format(
            self.hostname, self.port, self.uuid, self.array_names[ArrayKeys.RAW],
            self.array_names[ArrayKeys.GT_LABELS])



from gunpowder.provider_spec import ProviderSpec
from gunpowder.batch import Batch
from gunpowder.batch_request import BatchRequest

from .batch_provider import BatchProvider


class MergeProvider(BatchProvider):
    '''Merges different providers::

        (a + b + c) + MergeProvider()

    will create a provider that combines the arrays and points offered by
    ``a``, ``b``, and ``c``. Array and point keys of ``a``, ``b``, and ``c`` should be
    the disjoint.
    '''
    def __init__(self):
        self.key_to_provider = {}

    def setup(self):
        assert len(self.get_upstream_providers()) > 0, "at least one batch provider needs to be added to the MergeProvider"
        # Only allow merging if no two upstream_providers have the same
        # array/points keys
        error_message = "Key {} provided by more than one upstream provider. Node MergeProvider only allows to merge " \
                        "providers with different keys."
        for provider in self.get_upstream_providers():
            for key, spec in provider.spec.items():
                assert self.spec is None or key not in self.spec, error_message.format(key)
                self.provides(key, spec)
                self.key_to_provider[key] = provider

    def provide(self, request):

        # create upstream requests
        upstream_requests = {}
        for key, spec in request.items():

            provider = self.key_to_provider[key]
            if provider not in upstream_requests:
                upstream_requests[provider] = BatchRequest()

            upstream_requests[provider][key] = spec

        # execute requests, merge batches
        merged_batch = Batch()
        for provider, upstream_request in upstream_requests.items():

            batch = provider.request_batch(upstream_request)
            for key, array in batch.arrays.items():
                merged_batch.arrays[key] = array
            for key, points in batch.points.items():
                merged_batch.points[key] = points

        return merged_batch





from .freezable import Freezable
from copy import deepcopy
from gunpowder.coordinate import Coordinate
from gunpowder.roi import Roi
import logging
import numpy as np

logger = logging.getLogger(__name__)

class Array(Freezable):
    '''A numpy array with a specification describing the data.

    Args:

        data (array-like):

            The data to be stored in the array. Will be converted to a numpy
            array, if necessary.

        spec (:class:`ArraySpec`, optional):

            A spec describing the data.

        attrs (``dict``, optional):

            Optional attributes to describe this array.
    '''

    def __init__(self, data, spec=None, attrs=None):

        self.spec = deepcopy(spec)
        self.data = np.asarray(data)
        self.attrs = attrs

        if attrs is None:
            self.attrs = {}

        if spec is not None and spec.roi is not None:
            for d in range(len(spec.voxel_size)):
                assert spec.voxel_size[d]*data.shape[-spec.roi.dims()+d] == spec.roi.get_shape()[d], \
                        "ROI %s does not align with voxel size %s * data shape %s"%(spec.roi, spec.voxel_size, data.shape)

        self.freeze()

    def crop(self, roi, copy=True):
        '''Create a cropped copy of this Array.

        Args:

            roi(:class:`Roi`):

                ROI in world units to crop to.

            copy(``bool``):

                Make a copy of the data (default).
        '''

        assert self.spec.roi.contains(roi), "Requested crop ROI (%s) doesn't fit in array (%s)"\
        %(roi, self.spec.roi)

        voxel_size = self.spec.voxel_size
        data_roi = (roi - self.spec.roi.get_offset())/voxel_size
        slices = data_roi.get_bounding_box()

        while len(slices) < len(self.data.shape):
            slices = (slice(None),) + slices

        data = self.data[slices]
        if copy:
            data = np.array(data)

        spec = deepcopy(self.spec)
        attrs = deepcopy(self.attrs)
        spec.roi = deepcopy(roi)
        return Array(data, spec, attrs)

class ArrayKey(Freezable):
    '''A key to identify arrays in requests, batches, and across nodes.

    Used as key in :class:`BatchRequest` and :class:`Batch` to retrieve array
    specs or arrays.

    Args:

        identifier (``string``):

            A unique, human readable identifier for this array key. Will be
            used in log messages and to look up arrays in requests and batches.
            Should be upper case (like ``RAW``, ``GT_LABELS``). The identifier
            is unique: Two array keys with the same identifier will refer to
            the same array.
    '''

    def __init__(self, identifier):
        self.identifier = identifier
        self.hash = hash(identifier)
        self.freeze()
        logger.debug("Registering array key %s", self)
        setattr(ArrayKeys, self.identifier, self)

    def __eq__(self, other):
        return hasattr(other, 'identifier') and self.identifier == other.identifier

    def __hash__(self):
        return self.hash

    def __repr__(self):
        return self.identifier

class ArrayKeys:
    '''Convenience access to all created :class:``ArrayKey``s. A key generated
    with::

        raw = ArrayKey('RAW')

    can be retrieved as::

        ArrayKeys.RAW
    '''
    pass



import copy
from .coordinate import Coordinate
from .freezable import Freezable

class ArraySpec(Freezable):
    '''Contains meta-information about an array. This is used by
    :class:`BatchProviders<BatchProvider>` to communicate the arrays they
    offer, as well as by :class:`Arrays<Array>` to describe the data they
    contain.

    Attributes:

        roi (:class:`Roi`):

            The region of interested represented by this array spec. Can be
            ``None`` for :class:`BatchProviders<BatchProvider>` that allow
            requests for arrays everywhere, but will always be set for array
            specs that are part of a :class:`Array`.

        voxel_size (:class:`Coordinate`):

            The size of the spatial axises in world units.

        interpolatable (``bool``):

            Whether the values of this array can be interpolated.

        dtype (``np.dtype``):

            The data type of the array.
    '''

    def __init__(self, roi=None, voxel_size=None, interpolatable=None, dtype=None):

        self.roi = roi
        self.voxel_size = None if voxel_size is None else Coordinate(voxel_size)
        self.interpolatable = interpolatable
        self.dtype = dtype

        self.freeze()

    def copy(self):
        '''Create a copy of this spec.'''
        return copy.deepcopy(self)

    def __eq__(self, other):

        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):

        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __repr__(self):
        r = ""
        r += "ROI: " + str(self.roi) + ", "
        r += "voxel size: " + str(self.voxel_size) + ", "
        r += "interpolatable: " + str(self.interpolatable) + ", "
        r += "dtype: " + str(self.dtype)
        return r



import logging
import numpy as np

from gunpowder.ext import tensorflow as tf
from gunpowder.nodes.generic_predict import GenericPredict
from gunpowder.array import ArrayKey, Array
from gunpowder.tensorflow.local_server import LocalServer

logger = logging.getLogger(__name__)

class Predict(GenericPredict):
    '''Tensorflow implementation of :class:`gunpowder.nodes.Predict`.

    Args:

        checkpoint (``string``):

            Basename of a tensorflow checkpoint storing the tensorflow graph
            and associated tensor values and metadata, as created by
            :class:`gunpowder.nodes.Train`, for example.

        inputs (``dict``, ``string`` -> :class:``ArrayKey``):

            Dictionary from the names of input tensors in the network to
            array keys.

        outputs (``dict``, ``string`` -> :class:``ArrayKey``):

            Dictionary from the names of output tensors in the network to array
            keys. New arrays will be generated by this node for each entry (if
            requested downstream).

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`, optional):

            Used to set the specs of generated arrays (``outputs``). This is
            useful to set the ``voxel_size``, for example, if they differ from
            the voxel size of the input arrays. Only fields that are not
            ``None`` in the given :class:`ArraySpec` will be used.

        graph: (``string``, optional):

            An optional path to a tensorflow computation graph that should be
            used for prediction. The checkpoint is used to restore the values
            of matching variable names in the graph. Note that the graph
            specified here can differ from the one associated to the
            checkpoint.
    '''

    def __init__(
            self,
            checkpoint,
            inputs,
            outputs,
            array_specs=None,
            graph=None):

        super(Predict, self).__init__(
            inputs,
            outputs,
            array_specs,
            spawn_subprocess=True)
        self.checkpoint = checkpoint
        self.meta_graph = graph
        self.session = None
        self.graph = None

    def start(self):

        target = LocalServer.get_target()
        logger.info("Initializing tf session, connecting to %s...", target)

        self.graph = tf.Graph()
        self.session = tf.Session(
            target=target,
            graph=self.graph)

        with self.graph.as_default():
            self.__read_checkpoint()

    def predict(self, batch, request):

        logger.debug("predicting in batch %i", batch.id)

        array_outputs = self.__collect_requested_outputs(request)
        inputs = self.__collect_provided_inputs(batch)

        # compute outputs
        outputs = self.session.run(array_outputs, feed_dict=inputs)

        for array_key in array_outputs:
            spec = self.spec[array_key].copy()
            spec.roi = request[array_key].roi
            batch.arrays[array_key] = Array(
                outputs[array_key],
                spec)

        logger.debug("predicted in batch %i", batch.id)

    def stop(self):

        if self.session is not None:
            self.session.close()
            self.graph = None
            self.session = None

    def __read_checkpoint(self):

        logger.info("Reading checkpoint...")

        # read the graph associated to the checkpoint
        if self.meta_graph is None:
            saver = tf.train.import_meta_graph(
                self.checkpoint + '.meta',
                clear_devices=True)
        # read alternative, custom graph
        else:
            saver = tf.train.import_meta_graph(
                    self.meta_graph,
                    clear_devices=True)

        # restore variables from checkpoint
        saver.restore(self.session, self.checkpoint)

    def __collect_requested_outputs(self, request):

        array_outputs = {}

        for output_name, array_key in self.outputs.items():
            if array_key in request:
                array_outputs[array_key] = output_name

        return array_outputs

    def __collect_provided_inputs(self, batch):

        inputs = {}

        for input_name, input_key in self.inputs.items():
            if isinstance(input_key, ArrayKey):
                if input_key in batch.arrays:
                    inputs[input_name] = batch.arrays[input_key].data
                else:
                    logger.warn("batch does not contain %s, input %s will not "
                                "be set", input_key, input_name)
            elif isinstance(input_key, np.ndarray):
                inputs[input_name] = input_key
            elif isinstance(input_key, str):
                inputs[input_name] = getattr(batch, input_key)
            else:
                raise Exception(
                    "Unknown network input key {}, can't be given to "
                    "network".format(input_key))

        return inputs



import logging
import traceback

from gunpowder.nodes.batch_provider import BatchProvider

logger = logging.getLogger(__name__)

class BatchProviderTree(BatchProvider):

    def __init__(self, inputs=None, output=None):
        self.inputs = inputs
        self.output = output
        self.initialized = False

    def setup(self):
        if not self.initialized:
            self.__rec_setup(self.output)
            self.initialized = True
        else:
            logger.warning("batch provider setup() called more than once")

    def internal_teardown(self):
        self.__rec_teardown(self.output)
        self.initialized = False

    def add_upstream_provider(self, batch_provider):
        for input in self.inputs:
            input.add_upstream_provider(batch_provider)

    def get_upstream_providers(self):
        upstream_providers = []
        for input in self.inputs:
            upstream_providers += input.get_upstream_providers()
        return upstream_providers

    @property
    def spec(self):
        return self.output.spec

    def provide(self, request):

        if not self.initialized:
            raise RuntimeError("You are requesting a batch from an uninitialized provider ('setup()' has not been called). Avoid this by using the 'gunpowder.build' context manager, which also takes care of tearing the provider down if it is not used anymore.")

        return self.output.request_batch(request)

    def __add__(self, batch_provider):

        assert isinstance(batch_provider, BatchProvider), "Can only add BatchProvider to provider trees."
        logger.debug("linking %s -> %s"%(type(self.output).__name__,type(batch_provider).__name__))

        batch_provider.add_upstream_provider(self.output)
        return BatchProviderTree(self.inputs, batch_provider)

    def __radd__(self, batch_providers):

        assert isinstance(batch_providers, tuple), "Don't know how to r-add anything but tuples."
        for batch_provider in batch_providers:
            assert isinstance(batch_provider, BatchProvider), "Can only add BatchProvider to provider trees."

        for batch_provider in batch_providers:
            logger.debug("linking %s -> %s"%(type(batch_provider).__name__,type(self).__name__))
            self.add_upstream_provider(batch_provider)

        return BatchProviderTree(list(batch_providers), self.output)

    def __rec_setup(self, provider):

        for upstream_provider in provider.get_upstream_providers():
            self.__rec_setup(upstream_provider)
        provider.setup()

    def __rec_teardown(self, provider):

        for upstream_provider in provider.get_upstream_providers():
            self.__rec_teardown(upstream_provider)

        try:
            provider.internal_teardown()
        except Exception as e:
            # don't stop on exceptions, try to tear down as much as possible of
            # the DAG
            logger.error("encountered exception during internal teardown: " + str(e))
            traceback.print_exc()

def batch_provider_add(self, batch_provider):

    assert isinstance(batch_provider, BatchProvider), "Can only add BatchProvider to batch providers."
    logger.debug("linking %s -> %s"%(type(self).__name__,type(batch_provider).__name__))

    batch_provider.add_upstream_provider(self)
    return BatchProviderTree([self], batch_provider)

def batch_provider_radd(self, batch_providers):

    assert isinstance(batch_providers, tuple), "Don't know how to r-add anything but tuples."
    for batch_provider in batch_providers:
        assert isinstance(batch_provider, BatchProvider), "Can only add BatchProvider to batch providers."

    for batch_provider in batch_providers:
        logger.debug("linking %s -> %s"%(type(batch_provider).__name__,type(self).__name__))
        self.add_upstream_provider(batch_provider)

    return BatchProviderTree(list(batch_providers), self)

BatchProvider.__add__ = batch_provider_add
BatchProvider.__radd__ = batch_provider_radd



import zarr

class ZarrFile():
    '''To be used as a context manager, similar to h5py.File.'''

    def __init__(self, filename, mode):
        self.filename = filename
        self.mode = mode

    def __enter__(self):
        return zarr.open(self.filename, mode=self.mode)

    def __exit__(self, *args):
        pass



import logging

logger = logging.getLogger(__name__)

class build(object):

    def __init__(self, batch_provider):
        self.batch_provider = batch_provider

    def __enter__(self):
        try:
            self.batch_provider.setup()
        except:
            logger.error("something went wrong during the setup of the pipeline, calling tear down")
            self.batch_provider.internal_teardown()
            logger.debug("tear down completed")
            raise
        return self.batch_provider

    def __exit__(self, type, value, traceback):
        logger.debug("leaving context, tearing down pipeline")
        self.batch_provider.internal_teardown()
        logger.debug("tear down completed")



import logging
import random
import numpy as np

from .batch_filter import BatchFilter
from gunpowder.coordinate import Coordinate

logger = logging.getLogger(__name__)

class SimpleAugment(BatchFilter):
    '''Randomly mirror and transpose all :class:`Arrays<Array>` and
    :class:`Points` in a batch.

    Args:

        mirror_only (``list`` of ``int``, optional):

            If set, only mirror between the given axes. This is useful to
            exclude channels that have a set direction, like time.

        transpose_only (``list`` of ``int``, optional):

            If set, only transpose between the given axes. This is useful to
            limit the transpose to axes with the same resolution or to exclude
            non-spatial dimensions.
    '''

    def __init__(self, mirror_only=None, transpose_only=None):

        self.mirror_only = mirror_only
        self.transpose_only = transpose_only
        self.mirror_mask = None
        self.dims = None
        self.transpose_dims = None

    def setup(self):

        self.dims = self.spec.get_total_roi().dims()

        if self.mirror_only is None:
            self.mirror_mask = [ True ]*self.dims
        else:
            self.mirror_mask = [ d in self.mirror_only for d in range(self.dims) ]

        if self.transpose_only is None:
            self.transpose_dims = list(range(self.dims))
        else:
            self.transpose_dims = self.transpose_only

    def prepare(self, request):

        self.total_roi = request.get_total_roi()

        self.mirror = [
            random.randint(0,1)
            if self.mirror_mask[d] else 0
            for d in range(self.dims)
        ]

        t = list(self.transpose_dims)
        random.shuffle(t)
        self.transpose = list(range(self.dims))
        for o, n in zip(self.transpose_dims, t):
            self.transpose[o] = n

        logger.debug("mirror = " + str(self.mirror))
        logger.debug("transpose = " + str(self.transpose))

        reverse_transpose = [0]*self.dims
        for d in range(self.dims):
            reverse_transpose[self.transpose[d]] = d

        logger.debug("downstream request = " + str(request))

        self.__transpose_request(request, reverse_transpose)
        self.__mirror_request(request, self.mirror)

        logger.debug("upstream request = " + str(request))

    def process(self, batch, request):

        mirror = tuple(
                slice(None, None, -1 if m else 1)
                for m in self.mirror
        )
        # arrays
        for (array_key, array) in batch.arrays.items():

            if array_key not in request:
                continue

            array.data = array.data[mirror]
            if self.transpose != (0, 1, 2):
                array.data = array.data.transpose(self.transpose)
        # points
        total_roi_offset = self.total_roi.get_offset()
        for (points_key, points) in batch.points.items():

            if points_key not in request:
                continue

            for loc_id, syn_point in points.data.items():
                # mirror
                location_in_total_offset = np.asarray(syn_point.location) - total_roi_offset
                syn_point.location = np.asarray([self.total_roi.get_end()[dim] - location_in_total_offset[dim]
                                                 if m else syn_point.location[dim] for dim, m in enumerate(self.mirror)])
                # transpose
                if self.transpose != (0, 1, 2):
                    syn_point.location = np.asarray([syn_point.location[self.transpose[d]] for d in range(self.dims)])

                # due to the mirroring, points at the lower boundary of the ROI
                # could fall on the upper one, which excludes them from the ROI
                if not points.spec.roi.contains(syn_point.location):
                    del points.data[loc_id]

        # arrays & points
        for collection_type in [batch.arrays, batch.points]:
            for (key, collector) in collection_type.items():
                if key not in request:
                    continue
                logger.debug("total ROI: %s"%self.total_roi)
                logger.debug("upstream %s ROI: %s"%(key, collector.spec.roi))
                self.__mirror_roi(collector.spec.roi, self.total_roi, self.mirror)
                logger.debug("mirrored %s ROI: %s"%(key,collector.spec.roi))
                self.__transpose_roi(collector.spec.roi, self.transpose)
                logger.debug("transposed %s ROI: %s"%(key,collector.spec.roi))


    def __mirror_request(self, request, mirror):

        for key, spec in request.items():
            self.__mirror_roi(spec.roi, self.total_roi, mirror)

    def __transpose_request(self, request, transpose):

        for key, spec in request.items():
            self.__transpose_roi(spec.roi, transpose)

    def __mirror_roi(self, roi, total_roi, mirror):

        total_roi_offset = total_roi.get_offset()
        total_roi_shape = total_roi.get_shape()

        roi_offset = roi.get_offset()
        roi_shape = roi.get_shape()

        roi_in_total_offset = roi_offset - total_roi_offset
        end_of_roi_in_total = roi_in_total_offset + roi_shape
        roi_in_total_offset_mirrored = total_roi_shape - end_of_roi_in_total
        roi_offset = Coordinate(
                total_roi_offset[d] + roi_in_total_offset_mirrored[d] if mirror[d] else roi_offset[d]
                for d in range(self.dims)
        )

        roi.set_offset(roi_offset)

    def __transpose_roi(self, roi, transpose):

        offset = roi.get_offset()
        shape = roi.get_shape()
        offset = tuple(offset[transpose[d]] for d in range(self.dims))
        shape = tuple(shape[transpose[d]] for d in range(self.dims))
        roi.set_offset(offset)
        roi.set_shape(shape)



import fractions
from gunpowder.coordinate import Coordinate
from gunpowder.points import PointsKey
from gunpowder.points_spec import PointsSpec
from gunpowder.array import ArrayKey
from gunpowder.array_spec import ArraySpec
from .freezable import Freezable

class ProviderSpec(Freezable):
    '''A collection of (possibly partial) :class:`ArraySpecs<ArraySpec>` and
    :class:`PointsSpecs<PointsSpec>` describing a
    :class:`BatchProvider's<BatchProvider>` offered arrays and points.

    This collection mimics a dictionary. Specs can be added with::

        provider_spec = ProviderSpec()
        provider_spec[array_key] = ArraySpec(...)
        provider_spec[points_key] = PointsSpec(...)

    Here, ``array_key`` and ``points_key`` are :class:`ArrayKey` and
    :class:`PointsKey`. The specs can be queried with::

        array_spec = provider_spec[array_key]
        points_spec = provider_spec[points_key]

    Furthermore, pairs of keys/values can be iterated over using
    ``provider_spec.items()``.

    To access only array or points specs, use the dictionaries
    ``provider_spec.array_specs`` or ``provider_spec.points_specs``,
    respectively.

    Args:

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`):

            Initial array specs.

        points_specs (``dict``, :class:`PointsKey` -> :class:`PointsSpec`):

            Initial points specs.

    Attributes:

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`):

            Contains all array specs contained in this provider spec.

        points_specs (``dict``, :class:`PointsKey` -> :class:`PointsSpec`):

            Contains all points specs contained in this provider spec.
    '''

    def __init__(self, array_specs=None, points_specs=None):

        self.array_specs = {}
        self.points_specs = {}
        self.freeze()

        # use __setitem__ instead of copying the dicts, this ensures type tests
        # are run
        if array_specs is not None:
            for key, spec in array_specs.items():
                self[key] = spec
        if points_specs is not None:
            for key, spec in points_specs.items():
                self[key] = spec


    def __setitem__(self, key, spec):

        if isinstance(spec, ArraySpec):
            assert isinstance(key, ArrayKey), ("Only a ArrayKey is "
                                                        "allowed as key for a "
                                                        "ArraySpec value.")
            self.array_specs[key] = spec.copy()

        elif isinstance(spec, PointsSpec):
            assert isinstance(key, PointsKey), ("Only a PointsKey is "
                                                        "allowed as key for a "
                                                        "PointsSpec value.")
            self.points_specs[key] = spec.copy()

        else:
            raise RuntimeError("Only ArraySpec or PointsSpec can be set in a "
                               "%s."%type(self).__name__)

    def __getitem__(self, key):

        if isinstance(key, ArrayKey):
            return self.array_specs[key]

        elif isinstance(key, PointsKey):
            return self.points_specs[key]

        else:
            raise RuntimeError(
                "Only ArrayKey or PointsKey can be used as keys in a "
                "%s."%type(self).__name__)

    def __len__(self):

        return len(self.array_specs) + len(self.points_specs)

    def __contains__(self, key):

        if isinstance(key, ArrayKey):
            return key in self.array_specs

        elif isinstance(key, PointsKey):
            return key in self.points_specs

        else:
            raise RuntimeError(
                "Only ArrayKey or PointsKey can be used as keys in a "
                "%s."%type(self).__name__)

    def __delitem__(self, key):

        if isinstance(key, ArrayKey):
            del self.array_specs[key]

        elif isinstance(key, PointsKey):
            del self.points_specs[key]

        else:
            raise RuntimeError(
                "Only ArrayKey or PointsKey can be used as keys in a "
                "%s."%type(self).__name__)

    def items(self):
        '''Provides a generator iterating over key/value pairs.'''

        for (k, v) in self.array_specs.items():
            yield k, v
        for (k, v) in self.points_specs.items():
            yield k, v

    def get_total_roi(self):
        '''Get the union of all the ROIs.'''

        total_roi = None
        for specs_type in [self.array_specs, self.points_specs]:
            for (_, spec) in specs_type.items():
                if total_roi is None:
                    total_roi = spec.roi
                else:
                    total_roi = total_roi.union(spec.roi)
        return total_roi

    def get_common_roi(self):
        '''Get the intersection of all the requested ROIs.'''

        common_roi = None
        for specs_type in [self.array_specs, self.points_specs]:
            for (_, spec) in specs_type.items():
                if common_roi is None:
                    common_roi = spec.roi
                else:
                    common_roi = common_roi.intersect(spec.roi)

        return common_roi

    def get_lcm_voxel_size(self, array_keys=None):
        '''Get the least common multiple of the voxel sizes in this spec.

        Args:

            array_keys (list of :class:`ArrayKey`, optional): If given,
                consider only the given array types.
        '''

        if array_keys is None:
            array_keys = self.array_specs.keys()

        if not array_keys:
            raise RuntimeError("Can not compute lcm voxel size -- there are "
                               "no array specs in this provider spec.")
        else:
            if not array_keys:
                raise RuntimeError("Can not compute lcm voxel size -- list of "
                                   "given array specs is empty.")

        lcm_voxel_size = None
        for key in array_keys:
            voxel_size = self.array_specs[key].voxel_size
            if lcm_voxel_size is None:
                lcm_voxel_size = voxel_size
            else:
                lcm_voxel_size = Coordinate(
                    (a * b // fractions.gcd(a, b)
                     for a, b in zip(lcm_voxel_size, voxel_size)))

        return lcm_voxel_size

    def __eq__(self, other):

        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):

        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __repr__(self):

        r = "\n"
        for (key, spec) in self.items():
            r += "\t%s: %s\n"%(key, spec)
        return r



import logging
import numpy as np

from .batch_filter import BatchFilter
from gunpowder.coordinate import Coordinate
from gunpowder.ext import malis
from gunpowder.array import Array

logger = logging.getLogger(__name__)

class AddAffinities(BatchFilter):
    '''Add an array with affinities for a given label array and neighborhood to 
    the batch. Affinity values are created one for each voxel and entry in the 
    neighborhood list, i.e., for each voxel and each neighbor of this voxel. 
    Values are 1 iff both labels (of the voxel and the neighbor) are equal and 
    non-zero.

    Args:

        affinity_neighborhood (``list`` of array-like):

            List of offsets for the affinities to consider for each voxel.

        labels (:class:`ArrayKey`):

            The array to read the labels from.

        affinities (:class:`ArrayKey`):

            The array to generate containing the affinities.

        labels_mask (:class:`ArrayKey`, optional):

            The array to use as a mask for ``labels``. Affinities connecting at
            least one masked out label will be masked out in
            ``affinities_mask``. If not given, ``affinities_mask`` will contain
            ones everywhere (if requested).

        unlabelled (:class:`ArrayKey`, optional):

            A binary array to indicate unlabelled areas with 0. Affinities from
            labelled to unlabelled voxels are set to 0, affinities between
            unlabelled voxels are masked out (they will not be used for
            training).

        affinities_mask (:class:`ArrayKey`, optional):

            The array to generate containing the affinitiy mask, as derived
            from parameter ``labels_mask``.
    '''

    def __init__(
            self,
            affinity_neighborhood,
            labels,
            affinities,
            labels_mask=None,
            unlabelled=None,
            affinities_mask=None):

        self.affinity_neighborhood = np.array(affinity_neighborhood)
        self.labels = labels
        self.unlabelled = unlabelled
        self.labels_mask = labels_mask
        self.affinities = affinities
        self.affinities_mask = affinities_mask

    def setup(self):

        assert self.labels in self.spec, (
            "Upstream does not provide %s needed by "
            "AddAffinities"%self.labels)

        voxel_size = self.spec[self.labels].voxel_size

        dims = self.affinity_neighborhood.shape[1]
        self.padding_neg = Coordinate(
                min([0] + [a[d] for a in self.affinity_neighborhood])
                for d in range(dims)
        )*voxel_size

        self.padding_pos = Coordinate(
                max([0] + [a[d] for a in self.affinity_neighborhood])
                for d in range(dims)
        )*voxel_size

        logger.debug("padding neg: " + str(self.padding_neg))
        logger.debug("padding pos: " + str(self.padding_pos))

        spec = self.spec[self.labels].copy()
        if spec.roi is not None:
            spec.roi = spec.roi.grow(self.padding_neg, -self.padding_pos)
        spec.dtype = np.float32

        self.provides(self.affinities, spec)
        if self.affinities_mask:
            self.provides(self.affinities_mask, spec)
        self.enable_autoskip()

    def prepare(self, request):

        if self.labels_mask:
            assert (
                request[self.labels].roi ==
                request[self.labels_mask].roi),(
                "requested GT label roi %s and GT label mask roi %s are not "
                "the same."%(
                    request[self.labels].roi,
                    request[self.labels_mask].roi))

        if self.unlabelled:
            assert (
                request[self.labels].roi ==
                request[self.unlabelled].roi),(
                "requested GT label roi %s and GT unlabelled mask roi %s are not "
                "the same."%(
                    request[self.labels].roi,
                    request[self.unlabelled].roi))

        if self.labels not in request:
            request[self.labels] = request[self.affinities].copy()

        labels_roi = request[self.labels].roi
        context_roi = request[self.affinities].roi.grow(
            -self.padding_neg,
            self.padding_pos)

        # grow labels ROI to accomodate padding
        labels_roi = labels_roi.union(context_roi)
        request[self.labels].roi = labels_roi

        # same for label mask
        if self.labels_mask and self.labels_mask in request:
            request[self.labels_mask].roi = \
                request[self.labels_mask].roi.union(context_roi)

        # and unlabelled mask
        if self.unlabelled and self.unlabelled in request:
            request[self.unlabelled].roi = \
                request[self.unlabelled].roi.union(context_roi)

        logger.debug("upstream %s request: "%self.labels + str(labels_roi))

    def process(self, batch, request):

        affinities_roi = request[self.affinities].roi

        logger.debug("computing ground-truth affinities from labels")
        affinities = malis.seg_to_affgraph(
                batch.arrays[self.labels].data.astype(np.int32),
                self.affinity_neighborhood
        ).astype(np.float32)


        # crop affinities to requested ROI
        offset = affinities_roi.get_offset()
        shift = -offset - self.padding_neg
        crop_roi = affinities_roi.shift(shift)
        crop_roi /= self.spec[self.labels].voxel_size
        crop = crop_roi.get_bounding_box()

        logger.debug("cropping with " + str(crop))
        affinities = affinities[(slice(None),)+crop]

        spec = self.spec[self.affinities].copy()
        spec.roi = affinities_roi
        batch.arrays[self.affinities] = Array(affinities, spec)

        if self.affinities_mask and self.affinities_mask in request:

            if self.labels_mask:

                logger.debug("computing ground-truth affinities mask from "
                             "labels mask")
                affinities_mask = malis.seg_to_affgraph(
                    batch.arrays[self.labels_mask].data.astype(np.int32),
                    self.affinity_neighborhood)
                affinities_mask = affinities_mask[(slice(None),)+crop]

            else:

                affinities_mask = np.ones_like(affinities)

            if self.unlabelled:

                # 1 for all affinities between unlabelled voxels
                unlabelled = (1 - batch.arrays[self.unlabelled].data)
                unlabelled_mask = malis.seg_to_affgraph(
                    unlabelled.astype(np.int32),
                    self.affinity_neighborhood)
                unlabelled_mask = unlabelled_mask[(slice(None),)+crop]

                # 0 for all affinities between unlabelled voxels
                unlabelled_mask = (1 - unlabelled_mask)

                # combine with mask
                affinities_mask = affinities_mask*unlabelled_mask

            affinities_mask = affinities_mask.astype(np.float32)
            batch.arrays[self.affinities_mask] = Array(affinities_mask, spec)

        else:

            if self.labels_mask is not None:
                logger.warning("GT labels does have a mask, but affinities "
                               "mask is not requested.")

        # crop labels to original label ROI
        if self.labels in request:
            roi = request[self.labels].roi
            batch.arrays[self.labels] = batch.arrays[self.labels].crop(roi)

        # same for label mask
        if self.labels_mask and self.labels_mask in request:
            roi = request[self.labels_mask].roi
            batch.arrays[self.labels_mask] = \
                batch.arrays[self.labels_mask].crop(roi)

        # and unlabelled mask
        if self.unlabelled and self.unlabelled in request:
            roi = request[self.unlabelled].roi
            batch.arrays[self.unlabelled] = \
                batch.arrays[self.unlabelled].crop(roi)

        batch.affinity_neighborhood = self.affinity_neighborhood



import logging
import numpy as np
import os

from .batch_filter import BatchFilter
from gunpowder.batch_request import BatchRequest
from gunpowder.ext import h5py

logger = logging.getLogger(__name__)

class Snapshot(BatchFilter):
    '''Save a passing batch in an HDF file.

    Args:

        dataset_names (``dict``, :class:`ArrayKey` -> ``string``):

            A dictionary from array keys to names of the datasets to store them
            in.

        output_dir (``string``):

            The directory to save the snapshots. Will be created, if it does
            not exist.

        output_filename (``string``):

            Template for output filenames. ``{id}`` in the string will be
            replaced with the ID of the batch. ``{iteration}`` with the training
            iteration (if training was performed on this batch).

        every (``int``):

            How often to save a batch. ``every=1`` indicates that every batch
            will be stored, ``every=2`` every second and so on. By default,
            every batch will be stored.

        additional_request (:class:`BatchRequest`):

            An additional batch request to merge with the passing request, if a
            snapshot is to be made. If not given, only the arrays that are in
            the batch anyway are recorded. This is useful to request additional
            arrays like loss gradients for visualization that are otherwise not
            needed.

        compression_type (``string`` or ``int``):

            Compression strategy.  Legal values are ``gzip``, ``szip``,
            ``lzf``. If an integer between 1 and 10, this indicates ``gzip``
            compression level.

        dataset_dtypes (``dict``, :class:`ArrayKey` -> data type):

            A dictionary from array keys to datatype (eg. ``np.int8``). If
            given, arrays are stored using this type. The original arrays
            within the pipeline remain unchanged.

        store_value_range (``bool``):

            If set to ``True`, store range of values in data set attributes.
        '''

    def __init__(
            self,
            dataset_names,
            output_dir='snapshots',
            output_filename='{id}.hdf',
            every=1,
            additional_request=None,
            compression_type=None,
            dataset_dtypes=None,
            store_value_range=False):
        self.dataset_names = dataset_names
        self.output_dir = output_dir
        self.output_filename = output_filename
        self.every = max(1,every)
        self.additional_request = BatchRequest() if additional_request is None else additional_request
        self.n = 0
        self.compression_type = compression_type
        self.store_value_range = store_value_range
        if dataset_dtypes is None:
            self.dataset_dtypes = {}
        else:
            self.dataset_dtypes = dataset_dtypes

    def prepare(self, request):

        self.record_snapshot = self.n%self.every == 0

        # append additional array requests, don't overwrite existing ones
        for array_key, spec in self.additional_request.array_specs.items():
            if array_key not in request.array_specs:
                request.array_specs[array_key] = spec

    def process(self, batch, request):

        if self.record_snapshot:

            try:
                os.makedirs(self.output_dir)
            except:
                pass

            snapshot_name = os.path.join(
                self.output_dir,
                self.output_filename.format(
                    id=str(batch.id).zfill(8),
                    iteration=int(batch.iteration or 0)))
            logger.info('saving to %s' %snapshot_name)
            with h5py.File(snapshot_name, 'w') as f:

                for (array_key, array) in batch.arrays.items():

                    if array_key not in self.dataset_names:
                        continue

                    ds_name = self.dataset_names[array_key]

                    if array_key in self.dataset_dtypes:
                        dtype = self.dataset_dtypes[array_key]
                        dataset = f.create_dataset(name=ds_name, data=array.data.astype(dtype), compression=self.compression_type)
                    else:
                        dataset = f.create_dataset(name=ds_name, data=array.data, compression=self.compression_type)
                    
                    if array.spec.roi is not None:
                        dataset.attrs['offset'] = array.spec.roi.get_offset()
                    dataset.attrs['resolution'] = self.spec[array_key].voxel_size

                    if self.store_value_range:
                        dataset.attrs['value_range'] = (
                            np.asscalar(array.data.min()),
                            np.asscalar(array.data.max()))

                    # if array has attributes, add them to the dataset
                    for attribute_name, attribute in array.attrs.items():
                        dataset.attrs[attribute_name] = attribute

                if batch.loss is not None:
                    f['/'].attrs['loss'] = batch.loss

        self.n += 1




import copy
import numpy as np

from .batch_provider import BatchProvider

class RandomProvider(BatchProvider):
    '''Randomly selects one of the upstream providers::

        (a + b + c) + RandomProvider()

    will create a provider that randomly relays requests to providers ``a``,
    ``b``, or ``c``. Array and point keys of ``a``, ``b``, and ``c`` should be
    the same.

    Args:
        probabilities (1-D array-like, optional): An optional list of
            probabilities for choosing upstream providers, given in the
            same order. Probabilities do not need to be normalized. Default
            is ``None``, corresponding to equal probabilities.
    '''

    def __init__(self, probabilities=None):
        self.probabilities = probabilities

        # automatically normalize probabilities to sum to 1
        if self.probabilities is not None:
            self.probabilities = [float(x)/np.sum(probabilities) for x in
                                  self.probabilities]

    def setup(self):

        assert len(self.get_upstream_providers()) > 0, "at least one batch provider needs to be added to the RandomProvider"
        if self.probabilities is not None:
            assert len(self.get_upstream_providers()) == len(
                self.probabilities), "if probabilities are specified, they " \
                                     "need to be given for each batch " \
                                     "provider added to the RandomProvider"

        common_spec = None

        # advertise outputs only if all upstream providers have them
        for provider in self.get_upstream_providers():

            if common_spec is None:
                common_spec = copy.deepcopy(provider.spec)
            else:
                for key, spec in provider.spec.items():
                    if key not in common_spec:
                        del common_spec[key]

        for key, spec in common_spec.items():
            self.provides(key, spec)

    def provide(self, request):
        return np.random.choice(self.get_upstream_providers(),
                                p=self.probabilities).request_batch(request)



import copy
from .points import PointsKey
from .points_spec import PointsSpec
from .provider_spec import ProviderSpec
from .roi import Roi
from .array import ArrayKey
from .array_spec import ArraySpec

class BatchRequest(ProviderSpec):
    '''A collection of (possibly partial) :class:`ArraySpec` and
    :class:`PointsSpec` forming a request.

    Inherits from :class:`ProviderSpec`.

    See :ref:`sec_requests_batches` for how to use a batch request to obtain a
    batch.
    '''

    def add(self, key, shape, voxel_size=None):
        '''Convenience method to add an array or point spec by providing only
        the shape of a ROI (in world units).

        A ROI with zero-offset will be generated. If more than one request is
        added, the ROIs with smaller shapes will be shifted to be centered in
        the largest one.

        Args:

            key (:class:`ArrayKey` or :class:`PointsKey`):

                The key for which to add a spec.

            shape (:class:`Coordinate`):

                A tuple containing the shape of the desired roi

            voxel_size (:class:`Coordinate`):

                A tuple contening the voxel sizes for each corresponding
                dimension
        '''

        if isinstance(key, ArrayKey):
            spec = ArraySpec()
        elif isinstance(key, PointsKey):
            spec = PointsSpec()
        else:
            raise RuntimeError("Only ArrayKey or PointsKey can be added.")

        spec.roi = Roi((0,)*len(shape), shape)

        if voxel_size is not None:
            spec.voxel_size = voxel_size

        self[key] = spec
        self.__center_rois()

    def copy(self):
        '''Create a copy of this request.'''
        return copy.deepcopy(self)

    def __center_rois(self):
        '''Ensure that all ROIs are centered around the same location.'''

        total_roi = self.get_total_roi()
        if total_roi is None:
            return

        center = total_roi.get_center()

        for specs_type in [self.array_specs, self.points_specs]:
            for key in specs_type:
                roi = specs_type[key].roi
                specs_type[key].roi = roi.shift(center - roi.get_center())



import logging
import random

from .batch_filter import BatchFilter
from gunpowder.profiling import Timing

logger = logging.getLogger(__name__)

class Reject(BatchFilter):
    '''Reject batches based on the masked-in vs. masked-out ratio.

    Args:

        mask (:class:`ArrayKey`):

            The mask to use.

        min_masked (``float``, optional):


            The minimal required ratio of masked-in vs. masked-out voxels.
            Defaults to 0.5.

        reject_probability (``float``, optional):

            The probability by which a batch that is not valid (less than
            min_masked) is actually rejected. Defaults to 1., i.e. strict
            rejection.
    '''

    def __init__(self, mask, min_masked=0.5, reject_probability=1.):

        self.mask = mask
        self.min_masked = min_masked
        self.reject_probability = reject_probability

    def setup(self):

        assert self.mask in self.spec, (
            "Reject can only be used if %s is provided"%self.mask)
        self.upstream_provider = self.get_upstream_provider()

    def provide(self, request):

        report_next_timeout = 10
        num_rejected = 0

        timing = Timing(self)
        timing.start()

        assert self.mask in request, (
            "Reject can only be used if a GT mask is requested")

        have_good_batch = False
        while not have_good_batch:

            batch = self.upstream_provider.request_batch(request)
            mask_ratio = batch.arrays[self.mask].data.mean()
            have_good_batch = mask_ratio>self.min_masked

            if not have_good_batch and self.reject_probability < 1.:
                have_good_batch = random.random() > self.reject_probability

            if not have_good_batch:

                logger.debug(
                    "reject batch with mask ratio %f at %s",
                    mask_ratio, batch.arrays[self.mask].spec.roi)
                num_rejected += 1

                if timing.elapsed() > report_next_timeout:

                    logger.warning(
                        "rejected %d batches, been waiting for a good one "
                        "since %ds", num_rejected, report_next_timeout)
                    report_next_timeout *= 2

            else:

                logger.debug(
                    "accepted batch with mask ratio %f at %s",
                    mask_ratio, batch.arrays[self.mask].spec.roi)

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch



import copy
import numpy as np
import time

from .freezable import Freezable

class Timing(Freezable):

    def __init__(self, node, method_name=None):
        self.__name = type(node).__name__
        self.__method_name = method_name
        self.__start = 0
        self.__first_start = 0
        self.__last_stop = 0
        self.__time = 0
        self.freeze()

    def start(self):
        self.__start = time.time()
        if self.__first_start == 0:
            self.__first_start = self.__start

    def stop(self):
        if self.__start == 0:
            return
        t = time.time()
        self.__time += (t - self.__start)
        self.__start = 0
        self.__last_stop = t

    def elapsed(self):
        '''Accumulated time elapsed between calls to start() and stop().'''

        if self.__start == 0:
            return self.__time

        return self.__time + (time.time() - self.__start)

    def span(self):
        '''Timestamps of the first call to start() and last call to stop().'''
        return self.__first_start, self.__last_stop

    def get_node_name(self):
        return self.__name

    def get_method_name(self):
        return self.__method_name

class TimingSummary(Freezable):
    '''Holds repeated Timings of the same node/method to be queried for statistics.'''

    def __init__(self):
        self.timings = []
        self.times = []
        self.freeze()

    def add(self, timing):
        '''Add a Timing to this summary.'''
        self.timings.append(timing)
        self.times.append(timing.elapsed())

    def merge(self, other):
        '''Merge another summary into this one.'''
        for timing in other.timings:
            self.add(timing)

    def counts(self):
        return len(self.times)

    def min(self):
        return np.min(self.times)

    def max(self):
        return np.max(self.times)

    def mean(self):
        return np.mean(self.times)

    def median(self):
        return np.median(self.times)

class ProfilingStats(Freezable):

    def __init__(self):
        self.__summaries = {}
        self.freeze()

    def add(self, timing):
        '''Add a Timing instance. Timings are grouped by their class and method names.'''

        node_name = timing.get_node_name()
        method_name = timing.get_method_name()
        id = (node_name, method_name)

        if id not in self.__summaries:
            self.__summaries[id] = TimingSummary()
        self.__summaries[id].add(copy.deepcopy(timing))

    def merge_with(self, other):
        '''Combine statitics of two ProfilingStats.'''

        for id, summary in other.__summaries.items():
            if id in self.__summaries:
                self.__summaries[id].merge(copy.deepcopy(summary))
            else:
                self.__summaries[id] = copy.deepcopy(summary)

    def get_timing_summaries(self):
        '''Get a dictionary (node_name,method_name) -> TimingSummary.'''
        return self.__summaries

    def get_timing_summary(self, node_name, method_name=None):
        '''Get a :class:`TimingSummary` for the given node and method name.'''

        if (node_name, method_name) not in self.__summaries:
            raise RuntimeError("No timing summary for node %s, method %s"%(node_name,method_name))

        return self.__summaries[(node_name,method_name)]

    def span(self):
        '''Timestamps of the first call to start() and last call to stop() over 
        all Timings added.'''

        spans = [t.span() for (_, summary) in self.__summaries.items() for t in summary.timings]
        first_start = min([span[0] for span in spans])
        last_stop = max([span[1] for span in spans])

        return first_start, last_stop

    def span_time(self):
        '''Time between the first call to start() and last call to stop() over 
        any timing.'''

        start, stop = self.span()
        return stop - start



import copy
import logging
import numpy as np

from gunpowder.batch import Batch
from gunpowder.coordinate import Coordinate
from gunpowder.ext import h5py
from gunpowder.points import PointsKeys, Points
from gunpowder.points_spec import PointsSpec
from gunpowder.profiling import Timing
from gunpowder.roi import Roi
from gunpowder.nodes.batch_provider import BatchProvider

from gunpowder.contrib.points import PreSynPoint, PostSynPoint

logger = logging.getLogger(__name__)

class Hdf5PointsSource(BatchProvider):
    '''An HDF5 data source for :class:``Points``. Currently only supports a
    specific case where points represent pre- and post-synaptic markers.

    Args:

        filename (string): The HDF5 file.

        datasets (dict): Dictionary of :class:``PointsKey`` -> dataset names
            that this source offers.

        rois (dict): Dictionary of :class:``PointsKey`` -> :class:``Roi`` to
            set the ROI for each point set provided by this source.

    '''

    def __init__(
            self,
            filename,
            datasets,
            rois):

        self.filename = filename
        self.datasets = datasets
        self.rois = rois

        self.ndims = None

    def setup(self):

        hdf_file = h5py.File(self.filename, 'r')

        for (points_key, ds_name) in self.datasets.items():

            if ds_name not in hdf_file:
                raise RuntimeError("%s not in %s"%(ds_name, self.filename))

            spec = PointsSpec()
            spec.roi = self.rois[points_key]

            self.provides(points_key, spec)

        hdf_file.close()

    def provide(self, request):

        timing = Timing(self)
        timing.start()

        batch = Batch()

        with h5py.File(self.filename, 'r') as hdf_file:

            # if pre and postsynaptic locations required, their id
            # SynapseLocation dictionaries should be created together s.t. ids
            # are unique and allow to find partner locations
            if PointsKeys.PRESYN in request.points_specs or PointsKeys.POSTSYN in request.points_specs:
                assert request.points_specs[PointsKeys.PRESYN].roi == request.points_specs[PointsKeys.POSTSYN].roi
                # Cremi specific, ROI offset corresponds to offset present in the
                # synapse location relative to the raw data.
                dataset_offset = self.spec[PointsKeys.PRESYN].roi.get_offset()
                presyn_points, postsyn_points = self.__get_syn_points(
                    roi=request.points_specs[PointsKeys.PRESYN].roi,
                    syn_file=hdf_file,
                    dataset_offset=dataset_offset)

            for (points_key, request_spec) in request.points_specs.items():

                logger.debug("Reading %s in %s...", points_key, request_spec.roi)
                id_to_point = {
                    PointsKeys.PRESYN: presyn_points,
                    PointsKeys.POSTSYN: postsyn_points}[points_key]

                points_spec = self.spec[points_key].copy()
                points_spec.roi = request_spec.roi
                batch.points[points_key] = Points(data=id_to_point, spec=points_spec)

        logger.debug("done")

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch

    def __get_syn_points(self, roi, syn_file, dataset_offset=None):
        presyn_points_dict, postsyn_points_dict = {}, {}
        presyn_node_ids  = syn_file['annotations/presynaptic_site/partners'][:, 0].tolist()
        postsyn_node_ids = syn_file['annotations/presynaptic_site/partners'][:, 1].tolist()

        for node_nr, node_id in enumerate(syn_file['annotations/ids']):
            location = syn_file['annotations/locations'][node_nr]
            if dataset_offset is not None:
                logging.debug('adding global offset to points %i %i %i' %(dataset_offset[0],
                                                                          dataset_offset[1], dataset_offset[2]))
                location += dataset_offset


            # cremi synapse locations are in physical space
            if roi.contains(Coordinate(location)):
                if node_id in presyn_node_ids:
                    kind = 'PreSyn'
                    assert syn_file['annotations/types'][node_nr] == 'presynaptic_site'
                    syn_id = int(np.where(presyn_node_ids == node_id)[0])
                    partner_node_id = postsyn_node_ids[syn_id]
                elif node_id in postsyn_node_ids:
                    kind = 'PostSyn'
                    assert syn_file['annotations/types'][node_nr] == 'postsynaptic_site'
                    syn_id = int(np.where(postsyn_node_ids == node_id)[0])
                    partner_node_id = presyn_node_ids[syn_id]
                else:
                    raise Exception('Node id neither pre- no post-synaptic')

                partners_ids = [int(partner_node_id)]
                location_id  = int(node_id)

                props = {}
                if node_id in syn_file['annotations/comments/target_ids']:
                    props = {'unsure': True}

                # create synpaseLocation & add to dict
                if kind == 'PreSyn':
                    syn_point = PreSynPoint(location=location, location_id=location_id,
                                         synapse_id=syn_id, partner_ids=partners_ids, props=props)
                    presyn_points_dict[int(node_id)] = copy.deepcopy(syn_point)
                elif kind == 'PostSyn':
                    syn_point = PostSynPoint(location=location, location_id=location_id,
                                         synapse_id=syn_id, partner_ids=partners_ids, props=props)
                    postsyn_points_dict[int(node_id)] = copy.deepcopy(syn_point)

        return presyn_points_dict, postsyn_points_dict


    def __repr__(self):

        return self.filename



import copy
import logging
import os

from gunpowder.caffe.net_io_wrapper import NetIoWrapper
from gunpowder.ext import caffe
from gunpowder.nodes.generic_predict import GenericPredict
from gunpowder.array import ArrayKey, Array


logger = logging.getLogger(__name__)

class Predict(GenericPredict):
    '''Augments a batch with network predictions.

    Args:

        prototxt (``string``):

            Filename of the network prototxt.

        weights (``string``):

            Filename of the network weights.

        inputs (``dict``, ``string`` -> :class:`ArrayKey`):

            Dictionary from the names of input layers in the network to array
            keys.

        outputs (``dict``, ``string`` -> :class:`ArrayKey`):

            Dictionary from the names of output layers in the network to
            array keys. New arrays will be generated by this node for each
            entry (if requested downstream).

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`, optional):

            Used to set the specs of generated arrays (``outputs``). This is
            useful to set the ``voxel_size``, for example, if they differ from
            the voxel size of the input arrays. Only fields that are not
            ``None`` in the given :class:`ArraySpec` will be used.

        use_gpu (``int`` or ``None``, optional):

            Which GPU to use. Set to ``None`` for CPU mode.
    '''

    def __init__(
            self,
            prototxt,
            weights,
            inputs,
            outputs,
            array_specs=None,
            use_gpu=None):

        super(Predict, self).__init__(
            inputs,
            outputs,
            array_specs,
            spawn_subprocess=True)
        for f in [prototxt, weights]:
            if not os.path.isfile(f):
                raise RuntimeError("%s does not exist"%f)
        self.prototxt = prototxt
        self.weights = weights
        self.inputs = inputs
        self.outputs = outputs
        self.use_gpu = use_gpu

    def start(self):

        logger.info("Initializing solver...")

        if self.use_gpu is not None:

            logger.debug("Predict process: using GPU %d"%self.use_gpu)
            caffe.enumerate_devices(False)
            caffe.set_devices((self.use_gpu,))
            caffe.set_mode_gpu()
            caffe.select_device(self.use_gpu, False)

        self.net = caffe.Net(self.prototxt, self.weights, caffe.TEST)
        self.net_io = NetIoWrapper(self.net, self.outputs.keys())

    def predict(self, batch, request):

        self.net_io.set_inputs({
            input_name: batch.arrays[array_key].data
            for input_name, array_key in self.inputs.items()
        })

        self.net.forward()
        output = self.net_io.get_outputs()

        for output_name, array_key in self.outputs.items():
            spec = self.spec[array_key].copy()
            spec.roi = request[array_key].roi
            batch.arrays[array_key] = Array(
                    output[output_name][0], # strip #batch dimension
                    spec)

        return batch



import numbers

class Coordinate(tuple):
    '''A ``tuple`` of integers.

    Allows the following element-wise operators: addition, subtraction,
    multiplication, division, absolute value, and negation. This allows to
    perform simple arithmetics with coordinates, e.g.::

        shape = Coordinate((2, 3, 4))
        voxel_size = Coordinate((10, 5, 1))
        size = shape*voxel_size # == Coordinate((20, 15, 4))
    '''

    def __new__(cls, array_like):
        return super(Coordinate, cls).__new__(
            cls,
            [
                int(x)
                if x is not None
                else None
                for x in array_like])

    def dims(self):
        return len(self)

    def __neg__(self):

        return Coordinate(
            -a
            if a is not None
            else None
            for a in self)

    def __abs__(self):

        return Coordinate(
            abs(a)
            if a is not None
            else None
            for a in self)

    def __add__(self, other):

        assert isinstance(other, tuple), "can only add Coordinate or tuples to Coordinate"
        assert self.dims() == len(other), "can only add Coordinate of equal dimensions"

        return Coordinate(
            a+b
            if a is not None and b is not None
            else None
            for a, b in zip(self, other))

    def __sub__(self, other):

        assert isinstance(other, tuple), "can only subtract Coordinate or tuples to Coordinate"
        assert self.dims() == len(other), "can only subtract Coordinate of equal dimensions"

        return Coordinate(
            a-b
            if a is not None and b is not None
            else None
            for a, b in zip(self, other))

    def __mul__(self, other):

        if isinstance(other, tuple):

            assert self.dims() == len(other), "can only multiply Coordinate of equal dimensions"

            return Coordinate(
                a*b
                if a is not None and b is not None
                else None
                for a,b in zip(self, other))

        elif isinstance(other, numbers.Number):

            return Coordinate(
                a*other
                if a is not None
                else None
                for a in self)

        else:

            raise TypeError("multiplication of Coordinate with type %s not supported" %type(other))

    def __div__(self, other):

        if isinstance(other, tuple):

            assert self.dims() == len(other), "can only divide Coordinate of equal dimensions"

            return Coordinate(
                a/b
                if a is not None and b is not None
                else None
                for a,b in zip(self, other))

        elif isinstance(other, numbers.Number):

            return Coordinate(
                a/other
                if a is not None
                else None
                for a in self)

        else:

            raise TypeError("division of Coordinate with type %s not supported" % type(other))

    def __truediv__(self, other):

        if isinstance(other, tuple):

            assert self.dims() == len(other), "can only divide Coordinate of equal dimensions"

            return Coordinate(
                a/b
                if a is not None and b is not None
                else None
                for a,b in zip(self, other))

        elif isinstance(other, numbers.Number):

            return Coordinate(
                a/other
                if a is not None
                else None
                for a in self)

        else:

            raise TypeError("division of Coordinate with type %s not supported" % type(other))

    def __floordiv__(self, other):

        if isinstance(other, tuple):

            assert self.dims() == len(other), "can only divide Coordinate of equal dimensions"

            return Coordinate(
                a//b
                if a is not None and b is not None
                else None
                for a,b in zip(self, other))

        elif isinstance(other, numbers.Number):

            return Coordinate(
                a//other
                if a is not None
                else None
                for a in self)

        else:

            raise TypeError("division of Coordinate with type %s not supported" % type(other))



from gunpowder.points import Point

class PreSynPoint(Point):
    def __init__(self, location, location_id, synapse_id, partner_ids, props=None):
        """ Presynaptic locations
        :param location:     ndarray, [zyx]
        :param location_id:  int, unique for every synapse location across pre and postsynaptic locations 
        :param synapse_id:   int, unique for every synapse(synaptic partners have the same synapse_id, but different location_ids)
        :param partner_ids:  list of ints, location ids of postsynaptic partners
        :param props:        dict, properties
        """
        Point.__init__(self, location=location)
        self.thaw()

        self.location_id  = location_id
        self.synapse_id   = synapse_id
        self.partner_ids  = partner_ids
        if props is None:
            self.props = {}
        else:
            self.props = props

        self.freeze()

class PostSynPoint(Point):
    def __init__(self, location, location_id, synapse_id, partner_ids, props=None):
        """
        :param location:     ndarray, [zyx]
        :param location_id:  int, unique for every synapse location across pre and postsynaptic locations 
        :param synapse_id:   int, unique for every synapse(synaptic partners have the same synapse_id, but different location_ids)
        :param partner_ids:  list of int, location id of presynaptic partner
        :param props:        dict, properties
        """
        Point.__init__(self, location=location)
        self.thaw()

        self.location_id  = location_id
        self.synapse_id   = synapse_id
        self.partner_ids  = partner_ids
        if props is None:
            self.props = {}
        else:
            self.props = props

        self.freeze()



import copy
import logging
import numpy as np
from scipy.ndimage.filters import gaussian_filter

from .batch_filter import BatchFilter
from gunpowder.array import Array
from gunpowder.array_spec import ArraySpec
from gunpowder.coordinate import Coordinate
from gunpowder.freezable import Freezable
from gunpowder.morphology import enlarge_binary_map, create_ball_kernel
from gunpowder.ndarray import replace
from gunpowder.points import PointsKeys
from gunpowder.points_spec import PointsSpec
from gunpowder.roi import Roi

logger = logging.getLogger(__name__)

class RasterizationSettings(Freezable):
    '''Data structure to store parameters for rasterization of points.

    Args:

        radius (``float`` or ``tuple`` of ``float``):

            The radius (for balls) or sigma (for peaks) in world units.

        mode (``string``):

            One of ``ball`` or ``peak``. If ``ball`` (the default), a ball with the
            given ``radius`` will be drawn. If ``peak``, the point will be
            rasterized as a peak with values :math:`\exp(-|x-p|^2/\sigma)` with
            sigma set by ``radius``.

        mask (:class:`ArrayKey`, optional):

            Used to mask the rasterization of points. The array is assumed to
            contain discrete labels. The object id at the specific point being
            rasterized is used to intersect the rasterization to keep it inside
            the specific object.

        inner_radius_fraction (``float``, optional):

            Only for mode ``ball``.

            If set, instead of a ball, a hollow sphere is rastered. The radius
            of the whole sphere corresponds to the radius specified with
            ``radius``. This parameter sets the radius of the hollow area, as a
            fraction of ``radius``.

        fg_value (``int``, optional):

            Only for mode ``ball``.

            The value to use to rasterize points, defaults to 1.

        bg_value (``int``, optional):

            Only for mode ``ball``.

            The value to use to for the background in the output array,
            defaults to 0.
    '''
    def __init__(
            self,
            radius,
            mode='ball',
            mask=None,
            inner_radius_fraction=None,
            fg_value=1,
            bg_value=0):

        radius = np.array([radius]).flatten().astype(np.float64)

        if inner_radius_fraction is not None:
            assert (
                inner_radius_fraction > 0.0 and
                inner_radius_fraction < 1.0), (
                    "Inner radius fraction has to be between (excluding) 0 and 1")

        self.radius = radius
        self.mode = mode
        self.mask = mask
        self.inner_radius_fraction = inner_radius_fraction
        self.fg_value = fg_value
        self.bg_value = bg_value
        self.freeze()

class RasterizePoints(BatchFilter):
    '''Draw points into a binary array as balls of a given radius.

    Args:

        points (:class:``PointsKeys``):
            The key of the points to rasterize.

        array (:class:``ArrayKey``):
            The key of the binary array to create.

        array_spec (:class:``ArraySpec``, optional):

            The spec of the array to create. Use this to set the datatype and
            voxel size.

        settings (:class:``RasterizationSettings``, optional):
            Which settings to use to rasterize the points.
    '''

    def __init__(self, points, array, array_spec=None, settings=None):

        self.points = points
        self.array = array
        if array_spec is None:
            self.array_spec = ArraySpec()
        else:
            self.array_spec = array_spec
        if settings is None:
            self.settings = RasterizationSettings(1)
        else:
            self.settings = settings

    def setup(self):

        points_roi = self.spec[self.points].roi

        if self.array_spec.voxel_size is None:
            self.array_spec.voxel_size = Coordinate((1,)*points_roi.dims())

        if self.array_spec.dtype is None:
            if self.settings.mode == 'ball':
                self.array_spec.dtype = np.uint8
            else:
                self.array_spec.dtype = np.float32

        self.array_spec.roi = points_roi.copy()
        self.provides(
            self.array,
            self.array_spec)

        self.enable_autoskip()

    def prepare(self, request):

        if self.settings.mode == 'ball':
            context = np.ceil(self.settings.radius).astype(np.int)
        elif self.settings.mode == 'peak':
            context = np.ceil(2*self.settings.radius).astype(np.int)
        else:
            raise RuntimeError('unknown raster mode %s'%self.settings.mode)

        dims = self.array_spec.roi.dims()
        if len(context) == 1:
            context = context.repeat(dims)

        # request points in a larger area to get rasterization from outside
        # points
        points_roi = request[self.array].roi.grow(
                Coordinate(context),
                Coordinate(context))

        # however, restrict the request to the points actually provided
        points_roi = points_roi.intersect(self.spec[self.points].roi)
        request[self.points] = PointsSpec(roi=points_roi)

        if self.settings.mask is not None:

            mask_voxel_size = self.spec[self.settings.mask].voxel_size
            assert self.spec[self.array].voxel_size == mask_voxel_size, (
                "Voxel size of mask and rasterized volume need to be equal")

            new_mask_roi = points_roi.snap_to_grid(mask_voxel_size)
            if self.settings.mask in request:
                request[self.settings.mask].roi = \
                    request[self.settings.mask].roi.union(new_mask_roi)
            else:
                request[self.settings.mask] = \
                    ArraySpec(roi=new_mask_roi)

    def process(self, batch, request):

        points = batch.points[self.points]
        mask = self.settings.mask
        voxel_size = self.spec[self.array].voxel_size

        # get roi used for creating the new array (points_roi does no
        # necessarily align with voxel size)
        enlarged_vol_roi = points.spec.roi.snap_to_grid(voxel_size)
        offset = enlarged_vol_roi.get_begin() / voxel_size
        shape = enlarged_vol_roi.get_shape() / voxel_size
        data_roi = Roi(offset, shape)

        logger.debug("Points in %s", points.spec.roi)
        for i, point in points.data.items():
            logger.debug("%d, %s", i, point.location)
        logger.debug("Data roi in voxels: %s", data_roi)
        logger.debug("Data roi in world units: %s", data_roi*voxel_size)

        if len(points.data.items()) == 0:
            # If there are no points at all, just create an empty matrix.
            rasterized_points_data = np.zeros(data_roi.get_shape(),
                                              dtype=self.spec[self.array].dtype)
        elif mask is not None:

            mask_array = batch.arrays[mask].crop(enlarged_vol_roi)
            # get those component labels in the mask, that contain points
            labels = []
            for i, point in points.data.items():
                v = Coordinate(point.location / voxel_size)
                v -= data_roi.get_begin()
                labels.append(mask_array.data[v])
            # Make list unique
            labels = list(set(labels))

            # zero label should be ignored
            if 0 in labels:
                labels.remove(0)

            if len(labels) == 0:
                logger.debug("Points and provided object mask do not overlap. No points to rasterize.")
                rasterized_points_data = np.zeros(data_roi.get_shape(),
                                                  dtype=self.spec[self.array].dtype)
            else:
                # create data for the whole points ROI, "or"ed together over
                # individual object masks
                rasterized_points_data = np.sum(
                    [
                        self.__rasterize(
                            points,
                            data_roi,
                            voxel_size,
                            self.spec[self.array].dtype,
                            self.settings,
                            Array(data=mask_array.data==label, spec=mask_array.spec))

                        for label in labels
                    ],
                    axis=0)

        else:

            # create data for the whole points ROI without mask
            rasterized_points_data = self.__rasterize(
                points,
                data_roi,
                voxel_size,
                self.spec[self.array].dtype,
                self.settings)

        # fix bg/fg labelling if requested
        if (self.settings.bg_value != 0 or
            self.settings.fg_value != 1):

            replaced = replace(
                rasterized_points_data,
                [0, 1],
                [self.settings.bg_value, self.settings.fg_value])
            rasterized_points_data = replaced.astype(self.spec[self.array].dtype)

        # create array and crop it to requested roi
        spec = self.spec[self.array].copy()
        spec.roi = data_roi*voxel_size
        rasterized_points = Array(
            data=rasterized_points_data,
            spec=spec)
        batch.arrays[self.array] = rasterized_points.crop(request[self.array].roi)

        # restore requested ROI of points
        if self.points in request:
            request_roi = request[self.points].roi
            points.spec.roi = request_roi
            for i, p in points.data.items():
                if not request_roi.contains(p.location):
                    del points.data[i]

        # restore requested mask
        if mask is not None:
            batch.arrays[mask] = batch.arrays[mask].crop(request[mask].roi)

    def __rasterize(self, points, data_roi, voxel_size, dtype, settings, mask_array=None):
        '''Rasterize 'points' into an array with the given 'voxel_size'''

        mask = mask_array.data if mask_array is not None else None

        logger.debug("Rasterizing points in %s", points.spec.roi)

        # prepare output array
        rasterized_points = np.zeros(data_roi.get_shape(), dtype=dtype)

        # Fast rasterization currently only implemented for mode ball without
        # inner radius set
        use_fast_rasterization = (
            settings.mode == 'ball' and
            settings.inner_radius_fraction is None
        )

        if use_fast_rasterization:

            dims = len(rasterized_points.shape)

            # get structuring element for mode ball
            ball_kernel = create_ball_kernel(settings.radius, voxel_size)
            radius_voxel = Coordinate(np.array(ball_kernel.shape)/2)
            data_roi_base = Roi(
                    offset=Coordinate((0,)*dims),
                    shape=Coordinate(rasterized_points.shape))
            kernel_roi_base = Roi(
                    offset=Coordinate((0,)*dims),
                    shape=Coordinate(ball_kernel.shape))

        # Rasterize volume either with single voxel or with defined struct elememt
        for point in points.data.values():

            # get the voxel coordinate, 'Coordinate' ensures integer
            v = Coordinate(point.location/voxel_size)

            # get the voxel coordinate relative to output array start
            v -= data_roi.get_begin()

            # skip points outside of mask
            if mask is not None and not mask[v]:
                continue

            logger.debug(
                "Rasterizing point %s at %s",
                point.location,
                point.location/voxel_size - data_roi.get_begin())

            if use_fast_rasterization:

                # Calculate where to crop the kernel mask and the rasterized array
                shifted_kernel = kernel_roi_base.shift(v - radius_voxel)
                shifted_data = data_roi_base.shift(-(v - radius_voxel))
                arr_crop = data_roi_base.intersect(shifted_kernel)
                kernel_crop = kernel_roi_base.intersect(shifted_data)
                arr_crop_ind = arr_crop.get_bounding_box()
                kernel_crop_ind = kernel_crop.get_bounding_box()

                rasterized_points[arr_crop_ind] = np.logical_or(
                    ball_kernel[kernel_crop_ind],
                    rasterized_points[arr_crop_ind])

            else:

                rasterized_points[v] = 1

        # grow points
        if not use_fast_rasterization:

            if settings.mode == 'ball':

                enlarge_binary_map(
                    rasterized_points,
                    settings.radius,
                    voxel_size,
                    1.0 - settings.inner_radius_fraction,
                    in_place=True)

            else:

                sigmas = settings.radius/voxel_size

                gaussian_filter(
                    rasterized_points,
                    sigmas,
                    output=rasterized_points,
                    mode='constant')

                # renormalize to have 1 be the highest value
                max_value = np.max(rasterized_points)
                if max_value > 0:
                    rasterized_points /= max_value

        if mask_array is not None:
            # use more efficient bitwise operation when possible
            if settings.mode == 'ball':
                rasterized_points &= mask
            else:
                rasterized_points *= mask

        return rasterized_points



import numpy as np
import skimage

from .batch_filter import BatchFilter

class NoiseAugment(BatchFilter):
    '''Add random noise to an array. Uses the scikit-image function skimage.util.random_noise.
    See scikit-image documentation for more information on arguments and additional kwargs.

    Args:

        array (:class:`ArrayKey`):

            The intensity array to modify. Should be of type float and within range [-1, 1] or [0, 1].

        mode (``string``):

            Type of noise to add, see scikit-image documentation.

        seed (``int``):

            Optionally set a random seed, see scikit-image documentation.

        clip (``bool``):

            Whether to preserve the image range (either [-1, 1] or [0, 1]) by clipping values in the end, see
            scikit-image documentation
    '''

    def __init__(self, array, mode='gaussian', seed=None, clip=True, **kwargs):
        self.array = array
        self.mode = mode
        self.seed = seed
        self.clip = clip
        self.kwargs = kwargs

    def process(self, batch, request):

        raw = batch.arrays[self.array]

        assert raw.data.dtype == np.float32 or raw.data.dtype == np.float64, "Noise augmentation requires float types for the raw array (not " + str(raw.data.dtype) + "). Consider using Normalize before."
        assert raw.data.min() >= -1 and raw.data.max() <= 1, "Noise augmentation expects raw values in [-1,1] or [0,1]. Consider using Normalize before."
        raw.data = skimage.util.random_noise(raw.data, mode=self.mode, seed=self.seed, clip=self.clip, **self.kwargs )



import copy
from .coordinate import Coordinate
from .freezable import Freezable
import numbers
import numpy as np

class Roi(Freezable):
    '''A rectangular region of interest, defined by an offset and a shape.

    Similar to :class:`Coordinate`, supports simple arithmetics, e.g.::

        roi = Roi((1, 1, 1), (10, 10, 10))
        voxel_size = Coordinate((10, 5, 1))
        scale_shift = roi*voxel_size + 1 # == Roi((11, 6, 2), (101, 51, 11))

    Args:

        offset (array-like of ``int``, optional):

            The starting point (inclusive) of the ROI. Can be ``None``
            (default) if the ROI only characterizes a shape.

        shape (array-like):

            The shape of the ROI. Entries can be ``None`` to indicate
            unboundedness. If ``None`` is passed instead of a tuple, all
            dimensions are set to ``None``, if the number of dimensions can be
            inferred from ``offset``.
    '''

    def __init__(self, offset=None, shape=None):

        self.__offset = None
        self.__shape = None
        self.freeze()

        self.set_shape(shape)
        if offset is not None:
            self.set_offset(offset)

    def set_offset(self, offset):

        self.__offset = Coordinate(offset)
        self.__consolidate_offset()

    def set_shape(self, shape):
        '''Set the shape of this ROI.

        Args:

            shape (array-like or ``None``):

                The new shape. Entries can be ``None`` to indicate
                unboundedness. If ``None`` is passed instead of a tuple, all
                dimensions are set to ``None``, if the number of dimensions can
                be inferred from an existing offset or previous shape.
        '''

        if shape is None:

            if self.__shape is not None:

                dims = self.__shape.dims()

            else:

                assert self.__offset is not None, (
                    "Can not infer dimension of ROI (there is no offset or "
                    "previous shape). Call set_shape with a tuple.")

                dims = self.__offset.dims()

            self.__shape = Coordinate((None,)*dims)

        else:

            self.__shape = Coordinate(shape)

        self.__consolidate_offset()

    def __consolidate_offset(self):
        '''Ensure that offsets for unbound dimensions are None.'''

        if self.__offset is not None:

            assert self.__offset.dims() == self.__shape.dims(), (
                "offset dimension %d != shape dimension %d"%(
                    self.__offset.dims(),
                    self.__shape.dims()))

            self.__offset = Coordinate((
                o
                if s is not None else None
                for o, s in zip(self.__offset, self.__shape)))

    def get_offset(self):
        return self.__offset

    def get_begin(self):
        '''Smallest coordinate inside ROI.'''
        return self.__offset

    def get_end(self):
        '''Smallest coordinate which is component-wise larger than any inside ROI.'''
        if not self.__shape:
            return self.__offset

        return self.__offset + self.__shape

    def get_shape(self):
        return self.__shape

    def get_center(self):

        return self.__offset + self.__shape/2

    def to_slices(self):
        '''Get a ``tuple`` of ``slice`` that represent this ROI and can be used
        to index arrays.'''

        if self.__offset is None:
            return None

        return tuple(
                slice(
                    int(self.__offset[d])
                    if self.__shape[d] is not None
                    else None,
                    int(self.__offset[d] + self.__shape[d])
                    if self.__shape[d] is not None
                    else None)
                for d in range(self.dims())
        )

    def get_bounding_box(self):
        return self.to_slices()

    def dims(self):
        '''The the number of dimensions of this ROI.'''

        if self.__shape is None:
            return 0
        return self.__shape.dims()

    def size(self):
        '''Get the volume of this ROI. Returns ``None`` if the ROI is
        unbounded.'''

        if self.unbounded():
            return None

        size = 1
        for d in self.__shape:
            size *= d
        return size

    def empty(self):
        '''Test if this ROI is empty.'''

        return self.size() == 0

    def unbounded(self):
        '''Test if this ROI is unbounded.'''

        return None in self.__shape

    def contains(self, other):
        '''Test if this ROI contains ``other``, which can be another
        :class:`Roi` or a :class:`Coordinate`.'''

        if isinstance(other, Roi):

            if other.empty():
                return True

            return (
                self.contains(other.get_begin())
                and
                self.contains(other.get_end() - (1,)*other.dims()))

        return all([
            (b is None or p is not None and p >= b)
            and
            (e is None or p is not None and p < e)
            for p, b, e in zip(other, self.get_begin(), self.get_end() )
        ])

    def intersects(self, other):
        '''Test if this ROI intersects with another :class:`Roi`.'''

        assert self.dims() == other.dims()

        if self.empty() or other.empty():
            return False

        # separated if at least one dimension is separated
        separated = any([
            # a dimension is separated if:
            # none of the shapes is unbounded
            (None not in [b1, b2, e1, e2])
            and
            (
                # either b1 starts after e2
                (b1 >= e2)
                or
                # or b2 starts after e1
                (b2 >= e1)
            )
            for b1, b2, e1, e2 in zip(
                self.get_begin(),
                other.get_begin(),
                self.get_end(),
                other.get_end())
        ])

        return not separated

    def intersect(self, other):
        '''Get the intersection of this ROI with another :class:`Roi`.'''

        if not self.intersects(other):
            return Roi(shape=(0,)*self.dims()) # empty ROI

        begin = Coordinate((
            self.__left_max(b1, b2)
            for b1, b2 in zip(self.get_begin(), other.get_begin())
        ))
        end = Coordinate((
            self.__right_min(e1, e2)
            for e1, e2 in zip(self.get_end(), other.get_end())
        ))

        return Roi(begin, end - begin)

    def union(self, other):
        '''Get the union of this ROI with another :class:`Roi`.'''

        begin = Coordinate((
            self.__left_min(b1, b2)
            for b1, b2 in zip(self.get_begin(), other.get_begin())
        ))
        end = Coordinate((
            self.__right_max(e1, e2)
            for e1, e2 in zip(self.get_end(), other.get_end())
        ))

        return Roi(begin, end - begin)

    def shift(self, by):
        '''Shift this ROI.'''

        return Roi(self.__offset + by, self.__shape)

    def snap_to_grid(self, voxel_size, mode='grow'):
        '''Align a ROI with a given voxel size.

        Args:

            voxel_size (:class:`Coordinate`):

                The voxel size of the grid to snap to.

            mode (string, optional):

                How to align the ROI if it is not a multiple of the voxel size.
                Available modes are 'grow', 'shrink', and 'closest'. Defaults to
                'grow'.
        '''

        begin_in_voxel_fractions = (
            np.asarray(self.get_begin(), dtype=np.float32)/
            np.asarray(voxel_size))
        end_in_voxel_fractions = (
            np.asarray(self.get_end(), dtype=np.float32)/
            np.asarray(voxel_size))

        if mode == 'closest':
            begin_in_voxel = np.round(begin_in_voxel_fractions)
            end_in_voxel = np.round(end_in_voxel_fractions)
        elif mode == 'grow':
            begin_in_voxel = np.floor(begin_in_voxel_fractions)
            end_in_voxel = np.ceil(end_in_voxel_fractions)
        elif mode == 'shrink':
            begin_in_voxel = np.ceil(begin_in_voxel_fractions)
            end_in_voxel = np.floor(end_in_voxel_fractions)
        else:
            assert False, 'Unknown mode %s for snap_to_grid'%mode

        return Roi(
            begin_in_voxel*voxel_size,
            (end_in_voxel - begin_in_voxel)*voxel_size)

    def grow(self, amount_neg, amount_pos):
        '''Grow a ROI by the given amounts in each direction:

        Args:

            amount_neg (:class:`Coordinate` or ``None``):

                Amount (per dimension) to grow into the negative direction.

            amount_pos (:class:`Coordinate` or ``None``):

                Amount (per dimension) to grow into the positive direction.
        '''

        if amount_neg is None:
            amount_neg = Coordinate((0,)*self.dims())
        if amount_pos is None:
            amount_pos = Coordinate((0,)*self.dims())

        assert len(amount_neg) == self.dims()
        assert len(amount_pos) == self.dims()

        offset = self.__offset - amount_neg
        shape = self.__shape + amount_neg + amount_pos

        return Roi(offset, shape)

    def copy(self):
        '''Create a copy of this ROI.'''
        return copy.deepcopy(self)

    def __left_min(self, x, y):

        # None is considered -inf

        if x is None or y is None:
            return None
        return min(x, y)

    def __left_max(self, x, y):

        # None is considered -inf

        if x is None:
            return y
        if y is None:
            return x
        return max(x, y)

    def __right_min(self, x, y):

        # None is considered +inf

        if x is None:
            return y
        if y is None:
            return x
        return min(x, y)

    def __right_max(self, x, y):

        # None is considered +inf

        if x is None or y is None:
            return None
        return max(x, y)

    def __add__(self, other):

        assert isinstance(other, tuple), "can only add Coordinate or tuples to Roi"
        return self.shift(other)

    def __sub__(self, other):

        assert isinstance(other, Coordinate), "can only subtract Coordinate from Roi"
        return self.shift(-other)

    def __mul__(self, other):

        assert isinstance(other, tuple) or isinstance(other, numbers.Number), "can only multiply with a number or tuple of numbers"
        return Roi(self.__offset*other, self.__shape*other)

    def __div__(self, other):

        assert isinstance(other, tuple) or isinstance(other, numbers.Number), "can only divide by a number or tuple of numbers"
        return Roi(self.__offset/other, self.__shape/other)

    def __truediv__(self, other):

        assert isinstance(other, tuple) or isinstance(other, numbers.Number), "can only divide by a number or tuple of numbers"
        return Roi(self.__offset/other, self.__shape/other)

    def __floordiv__(self, other):

        assert isinstance(other, tuple) or isinstance(other, numbers.Number), "can only divide by a number or tuple of numbers"
        return Roi(self.__offset//other, self.__shape//other)

    def __eq__(self, other):

        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):

        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __repr__(self):
        if self.empty():
            return "[empty ROI]"
        slices = ", ".join(
            [
                (str(b) if b is not None else "") +
                ":" +
                (str(e) if e is not None else "")
                for b, e in zip(self.get_begin(), self.get_end())
            ])
        dims = ", ".join(
            str(a) if a is not None else "inf"
            for a in self.__shape
        )
        return "[" + slices + "] (" + dims + ")"



import copy
import logging
import multiprocessing

from .batch_filter import BatchFilter
from gunpowder.profiling import Timing
from gunpowder.producer_pool import ProducerPool

logger = logging.getLogger(__name__)

class WorkersDiedException(Exception):
    pass

class PreCache(BatchFilter):
    '''Pre-cache repeated equal batch requests. For the first of a series of
    equal batch request, a set of workers is spawned to pre-cache the batches
    in parallel processes. This way, subsequent requests can be served quickly.

    This node only makes sense if:

    1. Incoming batch requests are repeatedly the same.
    2. There is a source of randomness in upstream nodes.

    Args:

        cache_size (``int``):

            How many batches to hold at most in the cache.

        num_workers (``int``):

            How many processes to spawn to fill the cache.
    '''

    def __init__(self, cache_size=50, num_workers=20):

        self.current_request = None
        self.workers = None
        self.cache_size = cache_size
        self.num_workers = num_workers

    def teardown(self):

        if self.workers is not None:
            self.workers.stop()

    def provide(self, request):

        timing = Timing(self)
        timing.start()

        if request != self.current_request:

            if self.workers is not None:
                logger.info("new request received, stopping current workers...")
                self.workers.stop()

            self.current_request = copy.deepcopy(request)

            logger.info("starting new set of workers...")
            self.workers = ProducerPool([ lambda i=i: self.__run_worker(i) for i in range(self.num_workers) ], queue_size=self.cache_size)
            self.workers.start()

        logger.debug("getting batch from queue...")
        batch = self.workers.get()

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch

    def __run_worker(self, i):

        return self.get_upstream_provider().request_batch(self.current_request)



import copy
from .freezable import Freezable

class PointsSpec(Freezable):
    '''Contains meta-information about a set of points. This is used by
    :class:`BatchProviders<BatchProvider>` to communicate the points they
    offer, as well as by :class:`Points` to describe the data they contain.

    Attributes:

        roi (:class:`Roi`):

            The region of interested represented by this set of points.
    '''

    def __init__(self, roi=None):

        self.roi = roi

        self.freeze()

    def copy(self):
        '''Create a copy of this spec.'''
        return copy.deepcopy(self)

    def __eq__(self, other):

        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):

        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __repr__(self):
        r = ""
        r += "ROI: " + str(self.roi)
        return r



import tensorflow as tf

def conv_pass(
        fmaps_in,
        kernel_size,
        num_fmaps,
        num_repetitions,
        activation='relu',
        name='conv_pass'):
    '''Create a convolution pass::

        f_in --> f_1 --> ... --> f_n

    where each ``-->`` is a convolution followed by a (non-linear) activation
    function and ``n`` ``num_repetitions``. Each convolution will decrease the
    size of the feature maps by ``kernel_size-1``.

    Args:

        f_in:

            The input tensor of shape ``(batch_size, channels, depth, height,
            width)`` or ``(batch_size, channels, height, width)``.

        kernel_size:

            Size of the kernel. Forwarded to the tensorflow convolution layer.

        num_fmaps:

            The number of feature maps to produce with each convolution.

        num_repetitions:

            How many convolutions to apply.

        activation:

            Which activation to use after a convolution. Accepts the name of any
            tensorflow activation function (e.g., ``relu`` for ``tf.nn.relu``).

    '''

    fmaps = fmaps_in
    if activation is not None:
        activation = getattr(tf.nn, activation)

    conv_layer = getattr(
        tf.layers,
        {2: 'conv2d', 3: 'conv3d'}[fmaps_in.get_shape().ndims - 2])

    for i in range(num_repetitions):
        fmaps = conv_layer(
            inputs=fmaps,
            filters=num_fmaps,
            kernel_size=kernel_size,
            padding='valid',
            data_format='channels_first',
            activation=activation,
            name=name + '_%i'%i)

    return fmaps

def downsample(fmaps_in, factors, name='down'):

    pooling_layer = getattr(
        tf.layers,
        {2: 'max_pooling2d', 3: 'max_pooling3d'}[fmaps_in.get_shape().ndims - 2])

    fmaps = pooling_layer(
        fmaps_in,
        pool_size=factors,
        strides=factors,
        padding='valid',
        data_format='channels_first',
        name=name)

    return fmaps

def upsample(fmaps_in, factors, num_fmaps, activation='relu', name='up'):

    if activation is not None:
        activation = getattr(tf.nn, activation)

    conv_trans_layer = getattr(
        tf.layers,
        {2: 'conv2d_transpose', 3: 'conv3d_transpose'}[fmaps_in.get_shape().ndims - 2])

    fmaps = conv_trans_layer(
        fmaps_in,
        filters=num_fmaps,
        kernel_size=factors,
        strides=factors,
        padding='valid',
        data_format='channels_first',
        activation=activation,
        name=name)

    return fmaps

def crop_spatial(fmaps_in, shape):
    '''Crop only the spacial dimensions to match shape.

    Args:

        fmaps_in:

            The input tensor.

        shape:

            A list (not a tensor) with the requested shape [_, _, z, y, x] or
            [_, _, y, x].
    '''

    in_shape = fmaps_in.get_shape().as_list()

    offset = [0, 0] + [(in_shape[i] - shape[i]) // 2 for i in range(2, len(shape))]
    size = in_shape[0:2] + shape[2:]

    fmaps = tf.slice(fmaps_in, offset, size)

    return fmaps

def unet(
        fmaps_in,
        num_fmaps,
        fmap_inc_factor,
        downsample_factors,
        activation='relu',
        layer=0):
    '''Create a 2D or 3D U-Net::

        f_in --> f_left --------------------------->> f_right--> f_out
                    |                                   ^
                    v                                   |
                 g_in --> g_left ------->> g_right --> g_out
                             |               ^
                             v               |
                                   ...

    where each ``-->`` is a convolution pass (see ``conv_pass``), each `-->>` a
    crop, and down and up arrows are max-pooling and transposed convolutions,
    respectively.

    The U-Net expects tensors to have shape ``(batch=1, channels, depth, height,
    width)`` for 3D or ``(batch=1, channels, height, width)`` for 2D.

    This U-Net performs only "valid" convolutions, i.e., sizes of the feature
    maps decrease after each convolution.

    Args:

        fmaps_in:

            The input tensor.

        num_fmaps:

            The number of feature maps in the first layer. This is also the
            number of output feature maps.

        fmap_inc_factor:

            By how much to multiply the number of feature maps between layers.
            If layer 0 has ``k`` feature maps, layer ``l`` will have
            ``k*fmap_inc_factor**l``.

        downsample_factors:

            List of lists ``[z, y, x]`` or ``[y, x]`` to use to down- and
            up-sample the feature maps between layers.

        activation:

            Which activation to use after a convolution. Accepts the name of any
            tensorflow activation function (e.g., ``relu`` for ``tf.nn.relu``).

        layer:

            Used internally to build the U-Net recursively.
    '''

    prefix = "    "*layer
    print(prefix + "Creating U-Net layer %i"%layer)
    print(prefix + "f_in: " + str(fmaps_in.shape))

    # convolve
    f_left = conv_pass(
        fmaps_in,
        kernel_size=3,
        num_fmaps=num_fmaps,
        num_repetitions=2,
        activation=activation,
        name='unet_layer_%i_left'%layer)

    # last layer does not recurse
    bottom_layer = (layer == len(downsample_factors))
    if bottom_layer:
        print(prefix + "bottom layer")
        print(prefix + "f_out: " + str(f_left.shape))
        return f_left

    # downsample
    g_in = downsample(
        f_left,
        downsample_factors[layer],
        'unet_down_%i_to_%i'%(layer, layer + 1))

    # recursive U-net
    g_out = unet(
        g_in,
        num_fmaps=num_fmaps*fmap_inc_factor,
        fmap_inc_factor=fmap_inc_factor,
        downsample_factors=downsample_factors,
        activation=activation,
        layer=layer+1)

    print(prefix + "g_out: " + str(g_out.shape))

    # upsample
    g_out_upsampled = upsample(
        g_out,
        downsample_factors[layer],
        num_fmaps,
        activation=activation,
        name='unet_up_%i_to_%i'%(layer + 1, layer))

    print(prefix + "g_out_upsampled: " + str(g_out_upsampled.shape))

    # copy-crop
    f_left_cropped = crop_spatial(f_left, g_out_upsampled.get_shape().as_list())

    print(prefix + "f_left_cropped: " + str(f_left_cropped.shape))

    # concatenate along channel dimension
    f_right = tf.concat([f_left_cropped, g_out_upsampled], 1)

    print(prefix + "f_right: " + str(f_right.shape))

    # convolve
    f_out = conv_pass(
        f_right,
        kernel_size=3,
        num_fmaps=num_fmaps,
        num_repetitions=2,
        name='unet_layer_%i_right'%layer)

    print(prefix + "f_out: " + str(f_out.shape))

    return f_out



from .freezable import Freezable
import logging
import numpy as np

logger = logging.getLogger(__name__)

class Points(Freezable):
    '''A list of :class:`Points<Point>` with a specification describing the
    data.

    Args:

        data (``dict``, ``int`` -> :class:`Point`):

            A dictionary of IDs mapping to :class:`Points<Point>`.

        spec (:class:`PointsSpec`):

            A spec describing the data.
    '''

    def __init__(self, data, spec):
        self.data = data
        self.spec = spec
        self.freeze()

    def __repr__(self):
        return "%s, %s"%(self.data, self.spec)

class Point(Freezable):
    '''A point with a location, as stored in :class:`Points`.

    Args:

        location (array-like of ``float``):

            The location of this point.
    '''

    def __init__(self, location):
        self.location = np.array(location, dtype=np.float32)
        self.freeze()

    def __repr__(self):
        return str(self.location)

    def copy(self):
        return Point(self.location)

class PointsKey(Freezable):
    '''A key to identify lists of points in requests, batches, and across
    nodes.

    Used as key in :class:`BatchRequest` and :class:`Batch` to retrieve specs
    or lists of points.

    Args:

        identifier (``string``):

            A unique, human readable identifier for this points key. Will be
            used in log messages and to look up points in requests and batches.
            Should be upper case (like ``CENTER_POINTS``). The identifier is
            unique: Two points keys with the same identifier will refer to the
            same points.
    '''

    def __init__(self, identifier):
        self.identifier = identifier
        self.hash = hash(identifier)
        self.freeze()
        logger.debug("Registering points type %s", self)
        setattr(PointsKeys, self.identifier, self)

    def __eq__(self, other):
        return hasattr(other, 'identifier') and self.identifier == other.identifier

    def __hash__(self):
        return self.hash

    def __repr__(self):
        return self.identifier

class PointsKeys:
    '''Convenience access to all created :class:`PointsKey`s. A key generated
    with::

        centers = PointsKey('CENTER_POINTS')

    can be retrieved as::

        PointsKeys.CENTER_POINTS
    '''
    pass



from __future__ import absolute_import

from . import nodes
from .nodes import *

from .batch import Batch
from .batch_provider_tree import *
from .batch_request import BatchRequest
from .build import build
from .coordinate import Coordinate
from .points import Points, Point, PointsKey, PointsKeys
from .points_spec import PointsSpec
from .producer_pool import ProducerPool
from .provider_spec import ProviderSpec
from .roi import Roi
from .array import Array, ArrayKey, ArrayKeys
from .array_spec import ArraySpec
import gunpowder.caffe
import gunpowder.tensorflow
import gunpowder.contrib
import gunpowder.zoo



from random import choice
import logging
import numpy as np

from gunpowder.coordinate import Coordinate
from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)

class SpecifiedLocation(BatchFilter):
    '''Choses a batch at a location from the list provided at init, making sure
    it is in the bounding box of the upstream provider.

    Locations should be given in world units.

    Locations will be chosen in order or at random from the list depending on the
    ``choose_randomly`` parameter.

    If a location requires a shift outside the bounding box of any upstream provider
    the module will skip that location with a warning.

    Args:

        locations (``list`` of locations):

            Locations to center batches around.

        choose_randomly (``bool``):

            Defines whether locations should be picked in order or at random
            from the list.

        extra_data (``list`` of array-like):

            A list of data that will be passed along with the arrays provided
            by this node. This data will be appended as an attribute to the
            dataset so it must be a data format compatible with hdf5.
    '''

    def __init__(self, locations, choose_randomly=False, extra_data=None):

        self.coordinates = locations
        self.choose_randomly = choose_randomly
        self.loc_i = 0
        self.upstream_spec = None
        self.upstream_roi = None
        self.specified_shift = None

        if extra_data is not None:
            assert len(extra_data) == len(locations),\
                "extra_data (%d) should match the length of specified locations (%d)"%(len(extra_data),\
                len(locations))

        self.extra_data = extra_data

    def setup(self):

        self.upstream_spec = self.get_upstream_provider().spec
        self.upstream_roi = self.upstream_spec.get_total_roi()

        if self.upstream_roi is None:
            raise RuntimeError("Can not draw random samples from a provider\
                that does not have a bounding box.")

        # clear bounding boxes of all provided arrays and points --
        # SpecifiedLocation does know its locations at setup (checks on the fly)
        for key, spec in self.spec.items():
            spec.roi.set_shape(None)
            self.updates(key, spec)

    def prepare(self, request):

        shift_roi = None

        for key, spec in request.items():
            request_roi = spec.roi
            if key in self.upstream_spec:
                provided_roi = self.upstream_spec[key].roi
            else:
                raise Exception(
                    "Requested %s, but upstream does not provide it."%key)
            key_shift_roi = provided_roi.shift(-request_roi.get_begin()).grow((0, 0, 0),
                                                    -request_roi.get_shape())

            if shift_roi is None:
                shift_roi = key_shift_roi
            else:
                shift_roi = shift_roi.intersect(key_shift_roi)

        logger.debug("valid shifts for request in " + str(shift_roi))

        # shift to center
        center_shift = spec.roi.get_shape()/2 + spec.roi.get_offset()

        self.specified_shift = self._get_next_shift(center_shift)

        # Set shift for all requests
        for specs_type in [request.array_specs, request.points_specs]:
            for (key, spec) in specs_type.items():
                roi = spec.roi.shift(self.specified_shift)
                specs_type[key].roi = roi

        logger.debug("{}'th shift selected: {}".format(self.loc_i, self.specified_shift))

    def process(self, batch, request):
        # reset ROIs to request
        for (array_key, spec) in request.array_specs.items():
            batch.arrays[array_key].spec.roi = spec.roi
            if self.extra_data is not None:
                batch.arrays[array_key].attrs['specified_location_extra_data'] =\
                 self.extra_data[self.loc_i]

        for (points_key, spec) in request.points_specs.items():
            batch.points[points_key].spec.roi = spec.roi

        # change shift point locations to lie within roi
        for points_key in request.points_specs.keys():
            for point_id, point in batch.points[points_key].data.items():
                batch.points[points_key].data[point_id].location -= self.specified_shift

    def _get_next_shift(self, center_shift):
        # gets next corrdinate from list

        if self.choose_randomly:
            next_shift = choice(self.coordinates) - center_shift
        else:
            next_shift = Coordinate(self.coordinates[self.loc_i] - center_shift)
            self.loc_i += 1
            if self.loc_i >= len(self.coordinates):
                self.loc_i = 0
                logger.warning('Ran out of specified locations, looping list')
        return next_shift



from .batch_filter import BatchFilter
from gunpowder.coordinate import Coordinate
from gunpowder.array import ArrayKey, Array
import logging
import numbers
import numpy as np

logger = logging.getLogger(__name__)

class DownSample(BatchFilter):
    '''Downsample arrays in a batch by given factors.

    Args:

        source (:class:`ArrayKey`):

            The key of the array to downsample.

        factor (``int`` or ``tuple`` of ``int``):

            The factor to downsample with.

        target (:class:`ArrayKey`):

            The key of the array to store the downsampled ``source``.
    '''

    def __init__(self, source, factor, target):

        assert isinstance(source, ArrayKey)
        assert isinstance(target, ArrayKey)
        assert (
            isinstance(factor, numbers.Number) or isinstance(factor, tuple)), (
            "Scaling factor should be a number or a tuple of numbers.")

        self.source = source
        self.factor = factor
        self.target = target

    def setup(self):

        spec = self.spec[self.source].copy()
        spec.voxel_size *= self.factor
        self.provides(self.target, spec)

    def prepare(self, request):

        if self.target not in request:
            return

        logger.debug("preparing downsampling of " + str(self.source))

        request_roi = request[self.target].roi
        logger.debug("request ROI is %s"%request_roi)

        # add or merge to batch request
        if self.source in request:
            request[self.source].roi = request[self.source].roi.union(request_roi)
            logger.debug(
                "merging with existing request to %s",
                request[self.source].roi)
        else:
            request[self.source].roi = request_roi
            logger.debug("adding as new request")

    def process(self, batch, request):

        if self.target not in request:
            return

        input_roi = batch.arrays[self.source].spec.roi
        request_roi = request[self.target].roi

        assert input_roi.contains(request_roi)

        # downsample
        if isinstance(self.factor, tuple):
            slices = tuple(
                slice(None, None, k)
                for k in self.factor)
        else:
            slices = tuple(
                slice(None, None, self.factor)
                for i in range(input_roi.dims()))

        logger.debug("downsampling %s with %s", self.source, slices)

        crop = batch.arrays[self.source].crop(request_roi)
        data = crop.data[slices]

        # create output array
        spec = self.spec[self.target].copy()
        spec.roi = request_roi
        batch.arrays[self.target] = Array(data, spec)

        # restore requested rois
        request_roi = request[self.source].roi

        if input_roi != request_roi:

            assert input_roi.contains(request_roi)

            logger.debug(
                "restoring original request roi %s of %s from %s",
                request_roi, self.source, input_roi)
            cropped = batch.arrays[self.source].crop(request_roi)
            batch.arrays[self.source] = cropped



import logging
import numpy as np

from gunpowder.batch import Batch
from gunpowder.coordinate import Coordinate
from gunpowder.profiling import Timing
from gunpowder.roi import Roi
from gunpowder.array import Array
from gunpowder.array_spec import ArraySpec
from .batch_provider import BatchProvider

logger = logging.getLogger(__name__)


class Hdf5LikeSource(BatchProvider):
    '''An HDF5-like data source.

    Provides arrays from datasets accessed with an h5py-like API for each array
    key given. If the attribute ``resolution`` is set in a dataset, it will be
    used as the array's ``voxel_size``. If the attribute ``offset`` is set in a
    dataset, it will be used as the offset of the :class:`Roi` for this array.
    It is assumed that the offset is given in world units.

    Args:

        filename (``string``):

            The input file.

        datasets (``dict``, :class:`ArrayKey` -> ``string``):

            Dictionary of array keys to dataset names that this source offers.

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`, optional):

            An optional dictionary of array keys to array specs to overwrite
            the array specs automatically determined from the data file. This
            is useful to set a missing ``voxel_size``, for example. Only fields
            that are not ``None`` in the given :class:`ArraySpec` will be used.
    '''
    def __init__(
            self,
            filename,
            datasets,
            array_specs=None):

        self.filename = filename
        self.datasets = datasets

        if array_specs is None:
            self.array_specs = {}
        else:
            self.array_specs = array_specs

        # number of spatial dimensions
        self.ndims = None

    def _open_file(self, filename):
        raise NotImplementedError('Only implemented in subclasses')

    def setup(self):
        with self._open_file(self.filename) as data_file:
            for (array_key, ds_name) in self.datasets.items():

                if ds_name not in data_file:
                    raise RuntimeError("%s not in %s" % (ds_name, self.filename))

                spec = self.__read_spec(array_key, data_file, ds_name)

                self.provides(array_key, spec)

    def provide(self, request):

        timing = Timing(self)
        timing.start()

        batch = Batch()

        with self._open_file(self.filename) as data_file:
            for (array_key, request_spec) in request.array_specs.items():
                logger.debug("Reading %s in %s...", array_key, request_spec.roi)

                voxel_size = self.spec[array_key].voxel_size

                # scale request roi to voxel units
                dataset_roi = request_spec.roi / voxel_size

                # shift request roi into dataset
                dataset_roi = dataset_roi - self.spec[array_key].roi.get_offset() / voxel_size

                # create array spec
                array_spec = self.spec[array_key].copy()
                array_spec.roi = request_spec.roi

                # add array to batch
                batch.arrays[array_key] = Array(
                    self.__read(data_file, self.datasets[array_key], dataset_roi),
                    array_spec)

        logger.debug("done")

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch

    def _get_voxel_size(self, dataset):
        try:
            return Coordinate(dataset.attrs['resolution'])
        except Exception:  # todo: make specific when z5py supports it
            return None

    def _get_offset(self, dataset):
        try:
            return Coordinate(dataset.attrs['offset'])
        except Exception:  # todo: make specific when z5py supports it
            return None

    def __read_spec(self, array_key, data_file, ds_name):

        dataset = data_file[ds_name]

        if array_key in self.array_specs:
            spec = self.array_specs[array_key].copy()
        else:
            spec = ArraySpec()

        if spec.voxel_size is None:
            voxel_size = self._get_voxel_size(dataset)
            if voxel_size is None:
                voxel_size = Coordinate((1,)*len(dataset.shape))
                logger.warning("WARNING: File %s does not contain resolution information "
                               "for %s (dataset %s), voxel size has been set to %s. This "
                               "might not be what you want.",
                               self.filename, array_key, ds_name, spec.voxel_size)
            spec.voxel_size = voxel_size

        self.ndims = len(spec.voxel_size)

        if spec.roi is None:
            offset = self._get_offset(dataset)
            if offset is None:
                offset = Coordinate((0,)*self.ndims)

            shape= Coordinate(dataset.shape[-self.ndims:])
            spec.roi = Roi(offset, shape*spec.voxel_size)

        if spec.dtype is not None:
            assert spec.dtype == dataset.dtype, ("dtype %s provided in array_specs for %s, "
                                                 "but differs from dataset %s dtype %s" %
                                                 (self.array_specs[array_key].dtype,
                                                  array_key, ds_name, dataset.dtype))
        else:
            spec.dtype = dataset.dtype

        if spec.interpolatable is None:
            spec.interpolatable = spec.dtype in [
                np.float,
                np.float32,
                np.float64,
                np.float128,
                np.uint8  # assuming this is not used for labels
            ]
            logger.warning("WARNING: You didn't set 'interpolatable' for %s "
                           "(dataset %s). Based on the dtype %s, it has been "
                           "set to %s. This might not be what you want.",
                           array_key, ds_name, spec.dtype,
                           spec.interpolatable)

        return spec

    def __read(self, data_file, ds_name, roi):
        c = len(data_file[ds_name].shape) - self.ndims
        return np.asarray(data_file[ds_name][(slice(None),)*c + roi.to_slices()])

    def __repr__(self):

        return self.filename



import copy
import logging
from gunpowder.coordinate import Coordinate
from gunpowder.points_spec import PointsSpec
from gunpowder.provider_spec import ProviderSpec
from gunpowder.array import ArrayKey
from gunpowder.array_spec import ArraySpec

logger = logging.getLogger(__name__)

class BatchProvider(object):
    '''Superclass for all nodes in a `gunpowder` graph.

    A :class:`BatchProvider` provides :class:`Batches<Batch>` containing
    :class:`Arrays<Array>` and/or :class:`Points`. The available data is
    specified in a :class:`ProviderSpec` instance, accessible via :attr:`spec`.

    To create a new node, subclass this class and implement (at least)
    :func:`setup` and :func:`provide`.

    A :class:`BatchProvider` can be linked to any number of other
    :class:`BatchProviders<BatchProvider>` upstream. If your node accepts
    exactly one upstream provider, consider subclassing :class:`BatchFilter`
    instead.
    '''

    def add_upstream_provider(self, provider):
        self.get_upstream_providers().append(provider)
        return provider

    def get_upstream_providers(self):
        if not hasattr(self, 'upstream_providers'):
            self.upstream_providers = []
        return self.upstream_providers

    def setup(self):
        '''To be implemented in subclasses.

        Called during initialization of the DAG. Callees can assume that all
        upstream providers are set up already.

        In setup, call :func:`provides` to announce the arrays and points
        provided by this node.
        '''
        raise NotImplementedError("Class %s does not implement 'setup'"%self.name())

    def teardown(self):
        '''To be implemented in subclasses.

        Called during destruction of the DAG. Subclasses should use this to
        stop worker processes, if they used some.
        '''
        pass

    def provides(self, key, spec):
        '''Introduce a new output provided by this :class:`BatchProvider`.

        Implementations should call this in their :func:`setup` method, which
        will be called when the pipeline is build.

        Args:

            key (:class:`ArrayKey` or :class:`PointsKey`):

                The array or point set key provided.

            spec (:class:`ArraySpec` or :class:`PointsSpec`):

                The spec of the array or point set provided.
        '''

        logger.debug("Current spec of %s:\n%s", self.name(), self.spec)

        if self.spec is None:
            self._spec = ProviderSpec()

        assert key not in self.spec, (
            "Node %s is trying to add spec for %s, but is already "
            "provided."%(type(self).__name__, key))

        self.spec[key] = copy.deepcopy(spec)
        self.provided_items.append(key)

        logger.debug("%s provides %s with spec %s", self.name(), key, spec)

    def _init_spec(self):
        if not hasattr(self, '_spec'):
            self._spec = None

    def internal_teardown(self):

        logger.debug("Resetting spec of %s", self.name())
        self._spec = None
        self._provided_items = []

        self.teardown()

    @property
    def spec(self):
        '''Get the :class:`ProviderSpec` of this :class:`BatchProvider`.

        Note that the spec is only available after the pipeline has been build.
        Before that, it is ``None``.
        '''
        self._init_spec()
        return self._spec

    @property
    def provided_items(self):
        '''Get a list of the keys provided by this :class:`BatchProvider`.

        This list is only available after the pipeline has been build. Before
        that, it is empty.
        '''

        if not hasattr(self, '_provided_items'):
            self._provided_items = []

        return self._provided_items

    def remove_provided(self, request):
        '''Remove keys from `request` that are provided by this
        :class:`BatchProvider`.
        '''

        for key in self.provided_items:
            if key in request:
                del request[key]

    def request_batch(self, request):
        '''Request a batch from this provider.

        Args:

            request (:class:`BatchRequest`):

                A request containing (possibly partial)
                :class:`ArraySpecs<ArraySpec>` and
                :class:`PointSpecs<PointsSpec>`.
        '''

        logger.debug("%s got request %s", self.name(), request)

        self.check_request_consistency(request)

        batch = self.provide(copy.deepcopy(request))

        self.check_batch_consistency(batch, request)

        logger.debug("%s provides %s", self.name(), batch)

        return batch

    def check_request_consistency(self, request):

        for (key, request_spec) in request.items():

            assert key in self.spec, "%s: Asked for %s which this node does not provide"%(self.name(), key)
            assert (
                isinstance(request_spec, ArraySpec) or
                isinstance(request_spec, PointsSpec)), ("spec for %s is of type"
                                                        "%s"%(
                                                            key,
                                                            type(request_spec)))

            provided_spec = self.spec[key]

            provided_roi = provided_spec.roi
            request_roi = request_spec.roi

            if provided_roi is not None:
                assert provided_roi.contains(request_roi), "%s: %s's ROI %s outside of my ROI %s"%(self.name(), key, request_roi, provided_roi)

            if isinstance(key, ArrayKey):

                if request_spec.voxel_size is not None:
                    assert provided_spec.voxel_size == request_spec.voxel_size, "%s: voxel size %s requested for %s, but this node provides %s"%(
                            self.name(),
                            request_spec.voxel_size,
                            key,
                            provided_spec.voxel_size)

                if request_roi is not None:
                    for d in range(request_roi.dims()):
                        assert request_roi.get_shape()[d]%provided_spec.voxel_size[d] == 0, \
                                "in request %s, dimension %d of request %s is not a multiple of voxel_size %d"%(
                                        request,
                                        d,
                                        key,
                                        provided_spec.voxel_size[d])

    def check_batch_consistency(self, batch, request):

        for (array_key, request_spec) in request.array_specs.items():

            assert array_key in batch.arrays, "%s requested, but %s did not provide it."%(array_key,self.name())
            array = batch.arrays[array_key]
            assert array.spec.roi == request_spec.roi, "%s ROI %s requested, but ROI %s provided by %s."%(
                    array_key,
                    request_spec.roi,
                    array.spec.roi,
                    self.name()
            )
            assert array.spec.voxel_size == self.spec[array_key].voxel_size, (
                "voxel size of %s announced, but %s "
                "delivered for %s"%(
                    self.spec[array_key].voxel_size,
                    array.spec.voxel_size,
                    array_key))
            # ensure that the spatial dimensions are the same (other dimensions 
            # on top are okay, e.g., for affinities)
            if request_spec.roi is not None:
                dims = request_spec.roi.dims()
                data_shape = Coordinate(array.data.shape[-dims:])
                voxel_size = self.spec[array_key].voxel_size
                assert data_shape == request_spec.roi.get_shape()/voxel_size, "%s ROI %s requested, but size of array is %s*%s=%s provided by %s."%(
                        array_key,
                        request_spec.roi,
                        data_shape,
                        voxel_size,
                        data_shape*voxel_size,
                        self.name()
                )

        for (points_key, request_spec) in request.points_specs.items():

            assert points_key in batch.points, "%s requested, but %s did not provide it."%(points_key,self.name())
            points = batch.points[points_key]
            assert points.spec.roi == request_spec.roi, "%s ROI %s requested, but ROI %s provided by %s."%(
                                            points_key,
                                            request_spec.roi,
                                            points.spec.roi,
                                            self.name())

            for _, point in points.data.items():
                assert points.spec.roi.contains(point.location), (
                    "points provided by %s with ROI %s contain point at %s"%(
                        self.name(), points.spec.roi, point.location))

    def provide(self, request):
        '''To be implemented in subclasses.

        This function takes a :class:`BatchRequest` and should return the
        corresponding :class:`Batch`.

        Args:

            request(:class:`BatchRequest`):

                The request to process.
        '''
        raise NotImplementedError("Class %s does not implement 'provide'"%self.name())

    def name(self):
        return type(self).__name__

    def __repr__(self):

        return self.name() + ", providing: " + str(self.spec)



from .hdf5like_write_base import Hdf5LikeWrite
from gunpowder.coordinate import Coordinate
from gunpowder.ext import z5py
import logging
import os
import traceback

logger = logging.getLogger(__name__)

class N5Write(Hdf5LikeWrite):
    '''Assemble arrays of passing batches in one N5 container. This is useful
    to store chunks produced by :class:`Scan` on disk without keeping the
    larger array in memory. The ROIs of the passing arrays will be used to
    determine the position where to store the data in the dataset.

    Args:

        dataset_names (``dict``, :class:`ArrayKey` -> ``string``):

            A dictionary from array keys to names of the datasets to store them
            in.

        output_dir (``string``):

            The directory to save the N5 container. Will be created, if it does
            not exist.

        output_filename (``string``):

            The output filename of the container. Will be created, if it does
            not exist, otherwise data is overwritten in the existing container.

        compression_type (``string`` or ``int``):

            Compression strategy.  Legal values are ``gzip``, ``szip``,
            ``lzf``. If an integer between 1 and 10, this indicates ``gzip``
            compression level.

        dataset_dtypes (``dict``, :class:`ArrayKey` -> data type):

            A dictionary from array keys to datatype (eg. ``np.int8``). If
            given, arrays are stored using this type. The original arrays
            within the pipeline remain unchanged.
    '''

    def _get_voxel_size(self, dataset):

        logger.debug('Voxel size being reversed to account for N5 using column-major ordering')
        try:
            return Coordinate(dataset.attrs['resolution'][::-1])
        except Exception:
            logger.error(traceback.format_exc())
            return None

    def _get_offset(self, dataset):

        logger.debug('Offset being reversed to account for N5 using column-major ordering')
        try:
            return Coordinate(dataset.attrs['offset'][::-1])
        except Exception:
            logger.error(traceback.format_exc())
            return None

    def _set_voxel_size(self, dataset, voxel_size):

        logger.debug('Voxel size being reversed to account for N5 using column-major ordering')
        dataset.attrs['resolution'] = voxel_size[::-1]

    def _set_offset(self, dataset, offset):

        logger.debug('Offset being reversed to account for N5 using column-major ordering')
        dataset.attrs['offset'] = offset[::-1]

    def _open_file(self, filename):
        return z5py.File(filename, use_zarr_format=False)




from setuptools import setup
import subprocess
try:
    import base_string as string_types
except ImportError:
    string_types = str

extras_require = {
    'tensorflow': ['tensorflow'],
}


dep_set = set()
for value in extras_require.values():
    if isinstance(value, string_types):
        dep_set.add(value)
    else:
        dep_set.update(value)

extras_require['full'] = list(dep_set)

subprocess.call('pip install git+https://github.com/funkey/augment#egg=augment'.split())

setup(
        name='gunpowder',
        version='0.3.1',
        description='Data loading DAG for Greentea.',
        url='https://github.com/funkey/gunpowder',
        author='Jan Funke',
        author_email='jfunke@iri.upc.edu',
        license='MIT',
        packages=[
            'gunpowder',
            'gunpowder.nodes',
            'gunpowder.caffe',
            'gunpowder.caffe.nodes',
            'gunpowder.tensorflow',
            'gunpowder.tensorflow.nodes',
            'gunpowder.contrib',
            'gunpowder.contrib.nodes',
            'gunpowder.ext',
            'gunpowder.zoo',
            'gunpowder.zoo.tensorflow'
        ],
        install_requires=[
            "numpy",
            "scipy",
            "h5py",
            "scikit-image",
            "requests"
        ],
        extras_require=extras_require,
)



import logging

from gunpowder.coordinate import Coordinate
from gunpowder.ext import z5py
from .hdf5like_source_base import Hdf5LikeSource


logger = logging.getLogger(__name__)


class N5Source(Hdf5LikeSource):
    '''An `N5 <https://github.com/saalfeldlab/n5>`_ data source.

    Provides arrays from N5 datasets. If the attribute ``resolution`` is set
    in a N5 dataset, it will be used as the array's ``voxel_size``. If the
    attribute ``offset`` is set in a dataset, it will be used as the offset of
    the :class:`Roi` for this array. It is assumed that the offset is given in
    world units.

    Args:

        filename (``string``):

            The N5 directory.

        datasets (``dict``, :class:`ArrayKey` -> ``string``):

            Dictionary of array keys to dataset names that this source offers.

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`, optional):

            An optional dictionary of array keys to array specs to overwrite
            the array specs automatically determined from the data file. This
            is useful to set a missing ``voxel_size``, for example. Only fields
            that are not ``None`` in the given :class:`ArraySpec` will be used.
    '''
    def _get_voxel_size(self, dataset):
        try:
            logger.debug('Voxel size being reversed to account for N5 using column-major ordering')
            return Coordinate(dataset.attrs['resolution'][::-1])
        except Exception:  # todo: make specific when z5py supports it
            return None

    def _get_offset(self, dataset):
        try:
            logger.debug('Offset being reversed to account for N5 using column-major ordering')
            return Coordinate(dataset.attrs['offset'][::-1])
        except Exception:  # todo: make specific when z5py supports it
            return None

    def _open_file(self, filename):
        return z5py.File(filename, use_zarr_format=False)



import logging
import numpy as np

from gunpowder.caffe.net_io_wrapper import NetIoWrapper
from gunpowder.ext import caffe
from gunpowder.nodes.generic_train import GenericTrain
from gunpowder.array import Array, ArrayKey

logger = logging.getLogger(__name__)

class Train(GenericTrain):
    '''Caffe implementation of :class:`gunpowder.nodes.Train`.

    Args:

        solver_parameters (:class:`SolverParameters`):

            Parameters of the solver to use for training, contains the network
            description as well.

        inputs (``dict``, ``string`` -> :class:`ArrayKey`):

            Dictionary from names of input layers in the network to
            array keys.

        outputs (``dict``, ``string`` -> :class:`ArrayKey`):

            Dictionary from the names of output layers in the network to array
            keys. New arrays will be generated by this node for each entry (if
            requested downstream).

        gradients (``dict``, ``string`` -> :class:`ArrayKey`):

            Dictionary from the names of output layers in the network to array
            keys. New arrays containing the gradient of an output with respect
            to the loss will be generated by this node for each entry (if
            requested downstream).

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`, optional):

            Used to set the array specs of generated arrays (see ``outputs``
            and ``gradients``). This is useful to set the ``voxel_size``, for
            example, if they differ from the voxel size of the input arrays.
            Only fields that are not ``None`` in the given :class:`ArraySpec`
            will be used.

        use_gpu (``int`` or ``None``, optional):

            Which GPU to use. Set to ``None`` for CPU mode.
    '''

    def __init__(
            self,
            solver_parameters,
            inputs,
            outputs,
            gradients,
            array_specs=None,
            use_gpu=None):

        super(Train, self).__init__(
            inputs,
            outputs,
            gradients,
            array_specs,
            spawn_subprocess=True)
        self.solver_parameters = solver_parameters
        self.use_gpu = use_gpu
        self.solver = None
        self.net_io = None

    def start(self):

        logger.info("Initializing solver...")

        if self.use_gpu is not None:

            logger.debug("Train process: using GPU %d", self.use_gpu)
            caffe.enumerate_devices(False)
            caffe.set_devices((self.use_gpu,))
            caffe.set_mode_gpu()
            caffe.select_device(self.use_gpu, False)

        self.solver = caffe.get_solver(self.solver_parameters)
        if self.solver_parameters.resume_from is not None:
            logger.debug(
                "Train process: restoring solver state from %s",
                self.solver_parameters.resume_from)
            self.solver.restore(self.solver_parameters.resume_from)

        names_net_outputs = self.outputs.keys() + self.gradients.keys()
        self.net_io = NetIoWrapper(self.solver.net, names_net_outputs)

    def train_step(self, batch, request):

        data = {}
        for input_name, network_input in self.inputs.items():
            if isinstance(network_input, ArrayKey):
                if network_input in batch.arrays:
                    data[input_name] = batch.arrays[network_input].data
                else:
                    logger.warn("batch does not contain %s, input %s will not "
                                "be set", network_input, input_name)
            elif isinstance(network_input, np.ndarray):
                data[input_name] = network_input
            elif isinstance(network_input, str):
                data[input_name] = getattr(batch, network_input)
            else:
                raise Exception(
                    "Unknown network input type {}, can't be given to "
                    "network".format(network_input))
        self.net_io.set_inputs(data)

        loss = self.solver.step(1)
        # self.__consistency_check()

        requested_outputs = {
            name: array_key
            for name, array_key in self.outputs.items()
            if array_key in request.array_specs }

        if requested_outputs:

            output = self.net_io.get_outputs()

            for output_name, array_key in requested_outputs.items():

                spec = self.spec[array_key].copy()
                spec.roi = request[array_key].roi
                batch.arrays[array_key] = Array(
                    output[output_name][0], # strip #batch dimension
                    spec)

        requested_gradients = {
            name: array_key
            for name, array_key in self.gradients.items()
            if array_key in request.array_specs }

        if requested_gradients:

            diffs = self.net_io.get_output_diffs()

            for output_name, array_key in requested_gradients.items():

                spec = self.spec[array_key].copy()
                spec.roi = request[array_key].roi
                batch.arrays[array_key] = Array(
                    diffs[output_name][0], # strip #batch dimension
                    spec)

        batch.loss = loss
        batch.iteration = self.solver.iter

    def __consistency_check(self):

        diffs = self.net_io.get_output_diffs()
        for k in diffs:
            assert not np.isnan(diffs[k]).any(), "Detected NaN in output diff " + k
